#!/usr/bin/env python
"""
@author Jesse Haviland
"""

from collections import namedtuple

import numpy as np
from roboticstoolbox.robot import Robot  # DHLink
from roboticstoolbox.robot.ETS import ETS
from roboticstoolbox.robot.DHLink import DHLink  # HACK
from spatialmath.base.argcheck import \
    getvector, isscalar, verifymatrix, getmatrix
from spatialmath import base
from spatialmath.base.transforms3d import tr2jac, trinv
from spatialmath.base.transformsNd import t2r
from spatialmath import SE3, Twist3
import spatialmath.base.symbolic as sym
# from scipy.optimize import minimize, Bounds
from ansitable import ANSITable, Column

from roboticstoolbox.robot.DHLink import _check_rne
from frne import init, frne, delete

iksol = namedtuple("IKsolution", "q, success, reason")


class DHRobot(Robot):
    """
    Class for robots defined using Denavit-Hartenberg notation

    :param L: List of links which define the robot
    :type L: list(n)
    :param name: Name of the robot
    :type name: str
    :param manufacturer: Manufacturer of the robot
    :type manufacturer: str
    :param base: Locaation of the base
    :type base: SE3
    :param tool: Location of the tool
    :type tool: SE3
    :param gravity: Gravitational acceleration vector
    :type gravity: ndarray(3)

    A concrete superclass for arm type robots defined using Denavit-Hartenberg
    notation, that represents a serial-link arm-type robot.  Each link and
    joint in the chain is described by a DHLink-class object using
    Denavit-Hartenberg parameters (standard or modified).

    .. note:: Link subclass elements passed in must be all standard, or all
          modified, DH parameters.

    :reference:

        - Robotics, Vision & Control, Chaps 7-9,
          P. Corke, Springer 2011.
        - Robot, Modeling & Control,
          M.Spong, S. Hutchinson & M. Vidyasagar, Wiley 2006.

    """

    def __init__(
            self,
            L,
            **kwargs):

        # Verify L
        if not isinstance(L, list):
            raise TypeError('The links L must be stored in a list.')

        links = []
        self._n = 0

        # TODO rewrite using extend
        for i in range(len(L)):
            if isinstance(L[i], DHLink):
                # got a link
                links.append(L[i])
                self._n += 1
                L[i].id = self._n
                L[i].name = f"link{self._n}"

            elif isinstance(L[i], DHRobot):
                # got a robot
                for j in range(L[i].n):
                    links.append(L[i].links[j])
                    self._n += 1
                    L[i].id = self._n
                    L[i].name = f"link{self._n}"
            else:
                raise TypeError("Input can be only DHLink or DHRobot")

        super().__init__(links, **kwargs)

        # Check the DH convention
        self._mdh = self.links[0].mdh
        if not all([link.mdh == self.mdh for link in self.links]):
            raise ValueError('Robot has mixed D&H link conventions')

        # rne parameters
        self._rne_ob = None

    def __str__(self):
        """
        Pretty prints the DH Model of the robot. Will output angles in degrees

        :return: Pretty print of the robot model
        :rtype: str
        """
        s = ""
        deg = 180 / np.pi

        def qstr(j, link):
            j += 1
            if link.flip:
                s = f"-q{j:d}"
            else:
                s = f" q{j:d}"

            if L.offset != 0:
                sign = "+" if L.offset > 0 else "-"
                offset = abs(L.offset)
                if link.isprismatic:
                    s += f" {sign} {offset:}"
                else:
                    s += f" {sign} {offset * deg:.3g}\u00b0"
            return s

        def angle(theta, fmt=None):
            if sym.issymbol(theta):   # pragma nocover
                return "<<red>>" + str(theta)
            else:
                if fmt is not None:
                    return fmt.format(theta * deg) + "\u00b0"
                else:
                    return str(theta * deg) + "\u00b0"

        has_qlim = any([link._qlim is not None for link in self])
        if has_qlim:
            qlim_columns = [
                Column("q⁻", headalign="^"), Column("q⁺", headalign="^"),
            ]
            qlim = self.qlim

        else:
            qlim_columns = []
        if self.mdh:
            # MDH format
            table = ANSITable(
                Column("aⱼ₋₁", headalign="^"),
                Column("⍺ⱼ₋₁", headalign="^"),
                Column("θⱼ", headalign="^"),
                Column("dⱼ", headalign="^"),
                *qlim_columns,
                border="thick"
                )
            for j, L in enumerate(self):
                if has_qlim:
                    if L.isprismatic:
                        ql = [qlim[0, j], qlim[1, j]]
                    else:
                        ql = [angle(qlim[k, j], "{:.1f}") for k in [0, 1]]
                else:
                    ql = []
                if L.isprismatic:
                    table.row(
                        L.a, angle(L.alpha), angle(L.theta), qstr(j, L), *ql)
                else:
                    table.row(
                        L.a, angle(L.alpha), qstr(j, L), L.d, *ql)
        else:
            # DH format
            table = ANSITable(
                Column("θⱼ", headalign="^", colalign="<"),
                Column("dⱼ", headalign="^"),
                Column("aⱼ", headalign="^"),
                Column("⍺ⱼ", headalign="^"),
                *qlim_columns,
                border="thick"
                )
            for j, L in enumerate(self):
                if has_qlim:
                    if L.isprismatic:
                        ql = [qlim[0, j], qlim[1, j]]
                    else:
                        ql = [angle(qlim[k, j], "{:.1f}") for k in [0, 1]]
                else:
                    ql = []
                if L.isprismatic:
                    table.row(
                        angle(L.theta), qstr(j, L), L.a, angle(L.alpha), *ql)
                else:
                    table.row(qstr(j, L), L.d, L.a, angle(L.alpha), *ql)

        s = str(table)

        # show tool and base
        if self._tool is not None or self._tool is not None:
            table = ANSITable(
                Column("", colalign=">"),
                Column("", colalign="<"),
                border="thin", header=False)
            if self._base is not None:
                table.row(
                    "base", self._base.printline(
                        orient="rpy/xyz", fmt="{:.2g}", file=None))
            if self._tool is not None:
                table.row(
                    "tool", self._tool.printline(
                        orient="rpy/xyz", fmt="{:.2g}", file=None))
            s += "\n" + str(table)

        # show named configurations
        s += self.configurations_str()

        return s

    def __add__(self, L):
        nlinks = []

        # TODO - Should I do a deep copy here a physically copy the DHLinks
        # and not just the references?
        # Copy DHLink references to new list
        for i in range(self.n):
            nlinks.append(self.links[i])

        if isinstance(L, DHLink):
            nlinks.append(L)
        elif isinstance(L, DHRobot):
            for j in range(L.n):
                nlinks.append(L.links[j])
        else:
            raise TypeError("Can only combine DHRobots with other "
                            "DHRobots or DHLinks")

        return DHRobot(
            nlinks,
            name=self.name,
            manufacturer=self.manufacturer,
            base=self.base,
            tool=self.tool,
            gravity=self.gravity)

    # def copy(self):
    #     """
    #     Copy a robot

    #     :return: A deepish copy of the robot
    #     :rtype: Robot subclass instance
    #     """

    #     L = [link.copy() for link in self]

    #     new = DHRobot(
    #         L,
    #         name=self.name,
    #         manufacturer=self.manufacturer,
    #         base=self.base,
    #         tool=self.tool,
    #         gravity=self.gravity)

    #     new.q = self.q
    #     new.qd = self.qd
    #     new.qdd = self.qdd

    #     return new

    @property
    def mdh(self):
        """
        Modified Denavit-Hartenberg status

        :return: whether robot is defined using modified Denavit-Hartenberg
            notation
        :rtype: bool

        Example:

        .. runblock:: pycon

            >>> import roboticstoolbox as rtb
            >>> puma = rtb.models.DH.Puma560()
            >>> puma.mdh
            >>> panda = rtb.models.DH.Panda()
            >>> panda.mdh

        """
        return self._mdh

    @property
    def d(self):
        r"""
        Link offset values

        :return: List of link offset values :math:`d_j`.
        :rtype: ndarray(n,)

        The following are equivalent::

                robot.links[j].d
                robot.d[j]

        Example:

        .. runblock:: pycon

            >>> import roboticstoolbox as rtb
            >>> robot = rtb.models.DH.Puma560()
            >>> robot.d
        """
        v = []
        for i in range(self.n):
            v.append(self.links[i].d)
        return v

    @property
    def a(self):
        r"""
        Link length values

        :return: List of link length values :math:`a_j`.
        :rtype: ndarray(n,)

        The following are equivalent::

                robot.links[j].a
                robot.a[j]

        Example:

        .. runblock:: pycon

            >>> import roboticstoolbox as rtb
            >>> robot = rtb.models.DH.Puma560()
            >>> robot.a
        """
        v = []
        for i in range(self.n):
            v.append(self.links[i].a)
        return v

    @property
    def theta(self):
        r"""
        Joint angle values

        :return: List of joint angle values :math:`\theta_j`.
        :rtype: ndarray(n,)

        The following are equivalent::

                robot.links[j].theta
                robot.theta[j]

        Example:

        .. runblock:: pycon

            >>> import roboticstoolbox as rtb
            >>> robot = rtb.models.DH.Puma560()
            >>> robot.theta
        """
        v = []
        for i in range(self.n):
            v.append(self.links[i].theta)
        return v

    @property
    def alpha(self):
        r"""
        Link twist values

        :return: List of link twist values :math:`\alpha_j`.
        :rtype: ndarray(n,)

        The following are equivalent::

                robot.links[j].alpha
                robot.alpha[j]

        Example:

        .. runblock:: pycon

            >>> import roboticstoolbox as rtb
            >>> robot = rtb.models.DH.Puma560()
            >>> robot.alpha
        """
        v = []
        for i in range(self.n):
            v.append(self.links[i].theta)
        return v

    @property
    def r(self):
        r"""
        Link centre of mass values

        :return: Array of link centre of mass values :math:`r_j`.
        :rtype: ndarray(3,n)

        Example:

        .. runblock:: pycon

            >>> import roboticstoolbox as rtb
            >>> robot = rtb.models.DH.Puma560()
            >>> robot.r
        """
        # TODO tidyup
        v = np.copy(self.links[0].r)
        for i in range(1, self.n):
            v = np.c_[v, self.links[i].r]
        return v

    @property
    def offset(self):
        r"""
        Joint offset values

        :return: List of joint offset values :math:`\bar{q}_j`.
        :rtype: ndarray(n,)

        Example:

        .. runblock:: pycon

            >>> import roboticstoolbox as rtb
            >>> robot = rtb.models.DH.Puma560()
            >>> robot.offset
        """
        v = []
        for i in range(self.n):
            v.append(self.links[i].offset)
        return v

    @property
    def reach(self):
        r"""
        Reach of the robot

        :return: Maximum reach of the robot
        :rtype: float

        A conservative estimate of the reach of the robot. It is computed as
        :math:`\sum_j |a_j| + |d_j|` where :math:`d_j` is taken as the maximum
        joint coordinate (``qlim``) if the joint is primsmatic.

        .. note::

            - This is the *length sum* referred to in Craig's book
            - Probably an overestimate of the actual reach
            - Used by numerical inverse kinematics to scale translational
              error.
            - For a prismatic joint, uses ``qlim`` if it is set

        .. warning:: Computed on the first access. If kinematic parameters
              subsequently change this will not be reflected.
        """
        if self._reach is None:
            d = 0
            for link in self:
                d += abs(link.a) + (link.d)
                if link.isprismatic and link.qlim is not None:
                    d += link.qlim[1]
            self._reach = d
        return self._reach

    def A(self, j, q=None):
        """
        Link forward kinematics

        :param j: Joints to compute link transform for
        :type j: int, 2-tuple
        :param q: The joint configuration of the robot (Optional,
            if not supplied will use the stored q values)
        :type q: float ndarray(1,n)
        :return T: The transform between link frames
        :rtype T: SE3

        - ``robot.A(j, q)`` transform between link frames {0} and {n}.  ``q``
          is a vector (n) of joint variables.
        - ``robot.A([j1, j2], q)`` as above between link frames {j1} and {j2}.
        - ``robot.A(j)`` as above except uses the stored q value of the
          robot object.

        .. note:: Base and tool transforms are not applied.

        """

        if isscalar(j):
            j0 = 0
            jn = int(j)
        else:
            j = getvector(j, 2)
            j0 = int(j[0])
            jn = int(j[1])

        jn += 1

        if jn > self.n:
            raise ValueError("The joints value out of range")

        q = getvector(q)

        T = SE3()
        for i in range(j0, jn):
            T *= self.links[i].A(q[i])

        return T

    def islimit(self, q=None):
        """
        Joint limit test

        :param q: The joint configuration of the robot (Optional,
            if not supplied will use the stored q values)
        :type q: ndarray(n
        :return v: is a vector of boolean values, one per joint, False if
            ``q[j]`` is within the joint limits, else True
        :rtype v: bool list

        - ``robot.islimit(q)`` is a list of boolean values indicating if the
          joint configuration ``q`` is in violation of the joint limits.

        - ``robot.jointlimit()`` as above except uses the stored q value of the
          robot object.

        Example:

        .. runblock:: pycon

            >>> import roboticstoolbox as rtb
            >>> robot = rtb.models.DH.Puma560()
            >>> robot.islimit([0, 0, -4, 4, 0, 0])

        """
        q = self._getq(q)

        return [link.islimit(qk) for (link, qk) in zip(self, q)]

    def isspherical(self):
        """
        Test for spherical wrist

        :return: True if spherical wrist
        :rtype: bool

        Tests if the robot has a spherical wrist, that is, the last 3 axes are
        revolute and their axes intersect at a point.

        .. runblock:: pycon

            >>> import roboticstoolbox as rtb
            >>> robot = rtb.models.DH.Puma560()
            >>> robot.isspherical()

        """
        if self.n < 3:
            return False

        L = self.links[self.n - 3:self.n]

        alpha = [-np.pi / 2, np.pi / 2]

        return L[0].a == 0 \
            and L[1].a == 0 \
            and L[1].d == 0 \
            and (
                (L[0].alpha == alpha[0] and L[1].alpha == alpha[1])
                or
                (L[0].alpha == alpha[1] and L[1].alpha == alpha[0])
            ) \
            and L[0].sigma == 0 \
            and L[1].sigma == 0 \
            and L[2].sigma == 0

    def isprismatic(self):
        """
        Identify prismatic joints

        :return: a list of bool variables, one per joint, true if
            the corresponding joint is prismatic, otherwise false.
        :rtype: list of bool

        ``robot.isprismatic()`` is a bool list identifying the prismatic joints
        within the robot

        Example:

        .. runblock:: pycon

            >>> import roboticstoolbox as rtb
            >>> puma = rtb.models.DH.Puma560()
            >>> puma.isprismatic()
            >>> stanford = rtb.models.DH.Stanford()
            >>> stanford.isprismatic()
        """
        return [link.isprismatic for link in self]

    def isrevolute(self):
        """
        Identify revolute joints

        :return: a list of bool variables, one per joint, true if
            the corresponding joint is revolute, otherwise false.
        :rtype: list of bool

        ``robot.isrevolute()`` is a bool list identifying the revolute joints
        within the robot

        Example:

        .. runblock:: pycon

            >>> import roboticstoolbox as rtb
            >>> puma = rtb.models.DH.Puma560()
            >>> puma.isprismatic()
            >>> stanford = rtb.models.DH.Stanford()
            >>> stanford.isrevolute()
        """

        p = []

        for i in range(self.n):
            p.append(self.links[i].isrevolute)

        return p

    def config(self):
        """
        Return the joint configuration string

        :return: joint configuration string
        :rtype: str

        A string with one letter per joint: ``R`` for a revolute
        joint, and ``P`` for a prismatic joint.

        Example:

        .. runblock:: pycon

            >>> import roboticstoolbox as rtb
            >>> puma = rtb.models.DH.Puma560()
            >>> puma.config()
            >>> stanford = rtb.models.DH.Stanford()
            >>> stanford.config()
        """
        return ''.join(['R' if L.isrevolute else 'P' for L in self])

    def todegrees(self, q=None):
        """
        Convert joint angles to degrees

        :param q: The joint configuration of the robot (Optional,
            if not supplied will use the stored q values)
        :type q: ndarray(n)
        :return: a vector of joint coordinates in degrees and metres
        :rtype: ndarray(n)

        ``robot.todegrees(q)`` converts joint coordinates ``q`` to degrees
        taking into account whether elements of ``q`` correspond to revolute
        or prismatic joints, ie. prismatic joint values are not converted.

        ``robot.todegrees()`` as above except uses the stored q value of the
        robot object.

        Example:

        .. runblock:: pycon

            >>> import roboticstoolbox as rtb
            >>> from math import pi
            >>> stanford = rtb.models.DH.Stanford()
            >>> stanford.todegrees([pi/4, pi/8, 2, -pi/4, pi/6, pi/3])
        """

        q = self._getq(q)
        revolute = self.isrevolute()

        return np.array([
            q[k] * 180.0 / np.pi if
            revolute[k] else q[k] for k in range(len(q))
        ])

    def toradians(self, q):
        """
        Convert joint angles to radians

        :param q: The joint configuration of the robot (Optional,
            if not supplied will use the stored q values)
        :type q: ndarray(n)
        :return: a vector of joint coordinates in radians and metres
        :rtype: ndarray(n)

        ``robot.toradians(q)`` converts joint coordinates ``q`` to radians
        taking into account whether elements of ``q`` correspond to revolute
        or prismatic joints, ie. prismatic joint values are not converted.

        ``robot.toradians()`` as above except uses the stored q value of the
        robot object.

        Example:

        .. runblock:: pycon

            >>> import roboticstoolbox as rtb
            >>> stanford = rtb.models.DH.Stanford()
            >>> stanford.toradians([10, 20, 2, 30, 40, 50])
        """

        q = self._getq(q)
        revolute = self.isrevolute()

        return np.array([
            q[k] * np.pi / 180.0
            if revolute[k] else q[k] for k in range(len(q))
        ])

    def dhunique(self):
        """
        Print the unique DH parameters

        Print a table showing all the non-zero DH parameters, and their
        values.  This is the minimum set of kinematic parameters required
        to describe the robot.
        """

        table = ANSITable(
            Column("param"),
            Column("value", headalign="^", colalign="<", fmt="{:.4g}"),
            border="thin")
        for j, link in enumerate(self):
            if link.isprismatic:
                if link.theta != 0:
                    table.row(f"θ{j}", link.theta)
            elif link.isrevolute:
                if link.d != 0:
                    table.row(f"d{j}", link.d)
            if link.a != 0:
                    table.row(f"a{j}", link.a)
            if link.alpha != 0:
                    table.row(f"⍺{j}", link.alpha)
        table.print()

    def twists(self, q=None):
        """
        Joint axes as  twists

        :param q: The joint configuration of the robot
        :type q: array_like (n)
        :return: a vector of Twist objects
        :rtype: float ndarray(n,)
        :return: Represents the pose of the tool
        :rtype: SE3 instance

        - ``tw, T0 = twists(q)`` calculates a vector of Twist objects (n) that
          represent the axes of the joints for the robot with joint coordinates
          ``q`` (n). Also returns T0 which is an SE3 object representing the
          pose of the tool.

        - ``tw, T0 = twists()`` as above but the joint coordinates are taken
          to be zero.

        Example:

        .. runblock:: pycon

            >>> import roboticstoolbox as rtb
            >>> robot = rtb.models.DH.Puma560()
            >>> tw, T0 = robot.twists(robot.qz)
            >>> tw
            >>> T0

        """

        if q is None:
            q = np.zeros((self.n))

        T = self.fkine_all(q)
        tw = Twist3.Alloc(self.n)
        if self.mdh:
            # MDH case
            for j, link in enumerate(self):
                if link.sigma == 0:
                    tw[j] = Twist3.Revolute(T[j].a, T[j].t)
                else:
                    tw[j] = Twist3.Prismatic(T[j].a)
        else:
            # DH case
            for j, link in enumerate(self):
                if j == 0:
                    # first link case
                    if link.sigma == 0:
                        # revolute
                        tw[j] = Twist3.Revolute([0, 0, 1], [0, 0, 0])
                    else:
                        tw[j] = Twist3.Prismatic([0, 0, 1])  # prismatic
                else:
                    # subsequent links
                    if link.sigma == 0:
                        tw[j] = Twist3.Revolute(T[j-1].a, T[j-1].t)  # revolute
                    else:
                        tw[j] = Twist3.Prismatic(T[j-1].a)  # prismatic

        return tw, T[-1]

    def ets(self):

        # optionally start with the base transform
        if self._base is None:
            ets = ETS()
        else:
            ets = ETS.SE3(self._base)

        # add the links
        for link in self:
            ets *= link.ets()

        # optionally add the base transform
        if self._tool is not None:
            ets *= ETS.SE3(self._tool)

        return ets

    def fkine(self, q=None):
        """
        Forward kinematics

        :param q: The joint configuration (Optional,
            if not supplied will use the stored q values).
        :type q: ndarray(n) or ndarray(m,n)
        :return: Forward kinematics as an SE(3) matrix
        :rtype: SE3 instance

        - ``robot.fkine(q)`` computes the forward kinematics for the robot at
          joint configuration ``q``.

        - ``robot.fkine()`` as above except uses the stored ``q`` value of the
          robot object.

        If q is a 2D array, the rows are interpreted as the generalized joint
        coordinates for a sequence of points along a trajectory. ``q[k,j]`` is
        the j'th joint coordinate for the k'th trajectory configuration, and
        the returned ``SE3`` instance contains ``n`` values.

        Example:

        .. runblock:: pycon

            >>> import roboticstoolbox as rtb
            >>> puma = rtb.models.DH.Puma560()
            >>> puma.fkine([0, 0, 0, 0, 0, 0])

        .. note::

            - The robot's base or tool transform, if present, are incorporated
              into the result.
            - Joint offsets, if defined, are added to ``q`` before the forward
              kinematics are computed.
        """
        q = self._getq(q)

        T = SE3.Empty()
        for qr in getmatrix(q, (None, self.n)):

            first = True
            for q, L in zip(qr, self.links):
                if first:
                    Tr = L.A(q)
                    first = False
                else:
                    Tr *= L.A(q)

            if self._base is not None:
                Tr = self._base * Tr
            if self._tool is not None:
                Tr = Tr * self._tool
            T.append(Tr)

        return T

    def fkine_all(self, q=None, old=True):
        """
        Forward kinematics for all link frames

        :param q: The joint configuration of the robot (Optional,
            if not supplied will use the stored q values).
        :type q: ndarray(n) or ndarray(m,n)
        :param old: "old" behaviour, defaults to True
        :type old: bool, optional
        :return: Forward kinematics as an SE(3) matrix
        :rtype: SE3 instance with ``n`` values

        - ``fkine_all(q)`` evaluates fkine for each joint within a robot and
          returns a sequence of link frame poses.

        - ``fkine_all()`` as above except uses the stored q value of the
          robot object.

        Example:

        .. runblock:: pycon

            >>> import roboticstoolbox as rtb
            >>> puma = rtb.models.DH.Puma560()
            >>> T = puma.fkine_all([0, 0, 0, 0, 0, 0])
            >>> len(T)

        .. note::
            - Old behaviour is to return a list of ``n`` frames {1} to {n}, but
              if ``old=False`` it returns ``n``+1 frames {0} to {n}, ie. it
              includes the base frame.
            - The robot's base or tool transform, if present, are incorporated
              into the result.
            - Joint offsets, if defined, are added to q before the forward
              kinematics are computed.
        """
        q = self._getq(q)

        if self._base is not None:
            Tj = self._base
        else:
            Tj = SE3()
        first = True
        Tall = Tj

        for q, L in zip(q, self.links):
            if first:

                Tj *= L.A(q)
                if old:
                    Tall = Tj
                else:
                    Tall.append(Tj)
                first = False
            else:
                Tj *= L.A(q)
                Tall.append(Tj)
        return Tall

    def jacobe(self, q=None):
        r"""
        Manipulator Jacobian in end-effector frame

        :param q: Joint coordinate vector
        :type q: ndarray(n)
        :return J: The manipulator Jacobian in the end-effector frame
        :rtype: ndarray(6,n)

        - ``robot.jacobe(q)`` is the manipulator Jacobian matrix which maps
          joint  velocity to end-effector spatial velocity.

        End-effector spatial velocity :math:`\nu = (v_x, v_y, v_z, \omega_x, \omega_y, \omega_z)^T`
        is related to joint velocity by :math:`{}^{E}\!\nu = \mathbf{J}_m(q) \dot{q}`.

        Example:

        .. runblock:: pycon

            >>> import roboticstoolbox as rtb
            >>> puma = rtb.models.DH.Puma560()
            >>> puma.jacobe([0, 0, 0, 0, 0, 0])

        .. note:: This is the **geometric Jacobian** as described in texts by
            Corke, Spong etal., Siciliano etal.  The end-effector velocity is
            described in terms of translational and angular velocity, not a 
            velocity twist as per the text by Lynch & Park.
        """  # noqa

        q = getvector(q, self.n)

        n = self.n
        L = self.links
        J = np.zeros((6, self.n), dtype=q.dtype)

        U = self.tool.A

        for j in range(n - 1, -1, -1):
            if self.mdh == 0:
                # standard DH convention
                U = L[j].A(q[j]).A @ U

            if not L[j].sigma:
                # revolute axis
                d = np.array([
                    -U[0, 0] * U[1, 3] + U[1, 0] * U[0, 3],
                    -U[0, 1] * U[1, 3] + U[1, 1] * U[0, 3],
                    -U[0, 2] * U[1, 3] + U[1, 2] * U[0, 3]
                ])
                delta = U[2, :3]   # nz oz az
            else:
                # prismatic axis
                d = U[2, :3]      # nz oz az
                delta = np.zeros((3,))

            J[:, j] = np.r_[d, delta]

            if self.mdh != 0:
                # modified DH convention
                U = L[j].A(q[j]).A @ U

        return J

    def jacob0(self, q=None, T=None):
        r"""
        Manipulator Jacobian in world frame

        :param q: Joint coordinate vector
        :type q: ndarray(n)
        :param T: Forward kinematics if known, SE(3 matrix)
        :type T: SE3 instance
        :return J: The manipulator Jacobian in the world frame
        :rtype: ndarray(6,n)

        - ``robot.jacob0(q)`` is the manipulator Jacobian matrix which maps
          joint velocity to end-effector spatial velocity.

        End-effector spatial velocity :math:`\nu = (v_x, v_y, v_z, \omega_x, \omega_y, \omega_z)^T`
        is related to joint velocity by :math:`{}^{0}\!\nu = \mathbf{J}_0(q) \dot{q}`.

        Example:

        .. runblock:: pycon

            >>> import roboticstoolbox as rtb
            >>> puma = rtb.models.DH.Puma560()
            >>> puma.jacob0([0, 0, 0, 0, 0, 0])

        .. note:: This is the **geometric Jacobian** as described in texts by
            Corke, Spong etal., Siciliano etal.  The end-effector velocity is
            described in terms of translational and angular velocity, not a 
            velocity twist as per the text by Lynch & Park.
        """  # noqa
        q = getvector(q, self.n)

        if T is None:
            T = self.fkine(q)

        return tr2jac(trinv(T.A)) @ self.jacobe(q)

    def jacob_dot(self, q=None, qd=None):
        """
        Derivative of Jacobian

        :param q: The joint configuration of the robot (Optional,
            if not supplied will use the stored q values).
        :type q: float ndarray(n)
        :param qd: The joint configuration of the robot (Optional,
            if not supplied will use the stored qd values).
        :type qd: ndarray(n)
        :return: The derivative of the manipulator Jacobian
        :rtype:  ndarray(n)

        ``robot.jacob_dot(q, qd)`` computes the product (6) of the temporal
        derivative of the manipulator Jacobian (in the world frame) and the
        joint rates :math:`\frac{d}{dt} \dot{\mathbf{J}}(q) \dot{q}`.

        .. note::
            - This term appears in the formulation for operational space
              control :math:`\ddot{x} = \mathbf{J}(q) \ddot{q} + \dot{\mathbf{J}}(q) \dot{q}`.
            - Written as per the reference and not very efficient.

        :references:
            - Fundamentals of Robotics Mechanical Systems (2nd ed)
              J. Angleles, Springer 2003.
            - A unified approach for motion and force control of robot
              manipulators: The operational space formulation
              O Khatib, IEEE Journal on Robotics and Automation, 1987.

        """  # noqa

        if q is None:
            q = self.q

        if qd is None:
            qd = self.qd

        q = getvector(q, self.n)
        qd = getvector(qd, self.n)

        # Using the notation of Angeles:
        #   [Q,a] ~ [R,t] the per link transformation
        #   P ~ R   the cumulative rotation t2r(Tj) in world frame
        #   e       the last column of P, the local frame z axis in world
        #           coordinates
        #   w       angular velocity in base frame
        #   ed      deriv of e
        #   r       is distance from final frame
        #   rd      deriv of r
        #   ud      ??

        Q = np.zeros((3, 3, self.n))
        a = np.zeros((3, self.n))
        P = np.zeros((3, 3, self.n))
        e = np.zeros((3, self.n))
        w = np.zeros((3, self.n))
        ed = np.zeros((3, self.n))
        rd = np.zeros((3, self.n))
        r = np.zeros((3, self.n))
        ud = np.zeros((3, self.n))
        v = np.zeros((6, self.n))

        for i in range(self.n):
            T = self.links[i].A(q[i])
            Q[:, :, i] = T.R
            a[:, i] = T.t

        P[:, :, 0] = Q[:, :, 0]
        e[:, 0] = [0, 0, 1]

        for i in range(1, self.n):
            P[:, :, i] = P[:, :, i - 1] @ Q[:, :, i]
            e[:, i] = P[:, 2, i]

        # Step 1
        w[:, 0] = qd[0] * e[:, 0]

        for i in range(self.n - 1):
            w[:, i + 1] = (
                qd[i + 1] *
                np.array([0, 0, 1]) +
                Q[:, :, i].T @ w[:, i])

        # Step 2
        ed[:, 0] = np.array([0, 0, 1])

        for i in range(1, self.n):
            ed[:, i] = np.cross(w[:, i], e[:, i])

        # Step 3
        rd[:, self.n - 1] = np.cross(w[:, self.n - 1], a[:, self.n - 1])

        for i in range(self.n - 2, -1, -1):
            rd[:, i] = np.cross(w[:, i], a[:, i]) + Q[:, :, i] @ rd[:, i + 1]

        r[:, self.n - 1] = a[:, self.n - 1]

        for i in range(self.n - 2, -1, -1):
            r[:, i] = a[:, i] + Q[:, :, i] @ r[:, i + 1]

        ud[:, 0] = np.cross(e[:, 0], rd[:, 0])

        for i in range(1, self.n):
            ud[:, i] = \
                np.cross(ed[:, i], r[:, i]) + np.cross(e[:, i], rd[:, i])

        # Step 4
        # Swap ud and ed
        v[:, self.n - 1] = \
            qd[self.n - 1] * np.r_[ud[:, self.n - 1], ed[:, self.n - 1]]

        for i in range(self.n - 2, -1, -1):
            Ui = np.r_[
                np.c_[Q[:, :, i], np.zeros((3, 3))],
                np.c_[np.zeros((3, 3)), Q[:, :, i]]]

            v[:, i] = (
                qd[i] *
                np.r_[ud[:, i], ed[:, i]] +
                Ui @ v[:, i + 1])

        Jdot = v[:, 0]

        return Jdot

# -------------------------------------------------------------------------- #

    def _init_rne(self):
        # Compress link data into a 1D array
        L = np.zeros(24 * self.n)

        for i in range(self.n):
            j = i * 24
            L[j] = self.links[i].alpha
            L[j + 1] = self.links[i].a
            L[j + 2] = self.links[i].theta
            L[j + 3] = self.links[i].d
            L[j + 4] = self.links[i].sigma
            L[j + 5] = self.links[i].offset
            L[j + 6] = self.links[i].m
            L[j + 7:j + 10] = self.links[i].r.flatten()
            L[j + 10:j + 19] = self.links[i].I.flatten()
            L[j + 19] = self.links[i].Jm
            L[j + 20] = self.links[i].G
            L[j + 21] = self.links[i].B
            L[j + 22:j + 24] = self.links[i].Tc.flatten()

        self._rne_ob = init(self.n, self.mdh, L, self.gravity)

    def delete_rne(self):
        """
        Frees the memory holding the robot object in c if the robot object
        has been initialised in c.
        """
        if self._rne_ob is not None:
            delete(self._rne_ob)
            self._dynchanged = False
            self._rne_ob = None

    @_check_rne
    def rne(self, q, qd=None, qdd=None, gravity=None, fext=None):
        r"""
        Inverse dynamics

        :param q: Joint coordinates
        :type q: ndarray(n)
        :param qd: Joint velocity
        :type qd: ndarray(n)
        :param qdd: The joint accelerations of the robot
        :type qdd: ndarray(n)
        :param gravity: Gravitational acceleration to override robot's gravity
            value
        :type gravity: ndarray(6)
        :param fext: Specify wrench acting on the end-effector
                     :math:`W=[F_x F_y F_z M_x M_y M_z]`
        :type fext: ndarray(6)

        ``tau = rne(q, qd, qdd, grav, fext)`` is the joint torque required for
        the robot to achieve the specified joint position ``q`` (1xn), velocity
        ``qd`` (1xn) and acceleration ``qdd`` (1xn), where n is the number of
        robot joints. ``fext`` describes the wrench acting on the end-effector

        Trajectory operation:
        If q, qd and qdd (mxn) are matrices with m cols representing a
        trajectory then tau (mxn) is a matrix with cols corresponding to each
        trajectory step.

        .. note::
            - The torque computed contains a contribution due to armature
              inertia and joint friction.
            - If a model has no dynamic parameters set the result is zero.

        :seealso: :func:`rne_python`
        """
        trajn = 1

        try:
            q = getvector(q, self.n, 'row')
            qd = getvector(qd, self.n, 'row')
            qdd = getvector(qdd, self.n, 'row')
        except ValueError:
            trajn = q.shape[0]
            verifymatrix(q, (trajn, self.n))
            verifymatrix(qd, (trajn, self.n))
            verifymatrix(qdd, (trajn, self.n))

        if gravity is None:
            gravity = self.gravity
        else:
            gravity = getvector(gravity, 3)

        # The c function doesn't handle base rotation, so we need to hack the
        # gravity vector instead
        gravity = self.base.R.T @ gravity

        if fext is None:
            fext = np.zeros(6)
        else:
            fext = getvector(fext, 6)

        tau = np.zeros((trajn, self.n))

        for i in range(trajn):
            tau[i, :] = frne(
                self._rne_ob, q[i, :], qd[i, :], qdd[i, :], gravity, fext)

        if trajn == 1:
            return tau[0, :]
        else:
            return tau

    def rne_python(
            self, Q, QD=None, QDD=None,
            gravity=None, fext=None, debug=False, basewrench=False):
        """
        Compute inverse dynamics via recursive Newton-Euler formulation

        :param Q: Joint coordinates
        :param QD: Joint velocity
        :param QDD: Joint acceleration
        :param gravity: gravitational acceleration, defaults to attribute
            of self
        :type gravity: array_like(3), optional
        :param fext: end-effector wrench, defaults to None
        :type fext: array-like(6), optional
        :param debug: print debug information to console, defaults to False
        :type debug: bool, optional
        :param basewrench: compute the base wrench, defaults to False
        :type basewrench: bool, optional
        :raises ValueError: for misshaped inputs
        :return: Joint force/torques
        :rtype: NumPy array

        Recursive Newton-Euler for standard Denavit-Hartenberg notation.

        - ``rne_dh(q, qd, qdd)`` where the arguments have shape (n,) where n is
          the number of robot joints.  The result has shape (n,).
        - ``rne_dh(q, qd, qdd)`` where the arguments have shape (m,n) where n
          is the number of robot joints and where m is the number of steps in
          the joint trajectory.  The result has shape (m,n).
        - ``rne_dh(p)`` where the input is a 1D array ``p`` = [q, qd, qdd] with
          shape (3n,), and the result has shape (n,).
        - ``rne_dh(p)`` where the input is a 2D array ``p`` = [q, qd, qdd] with
          shape (m,3n) and the result has shape (m,n).

        .. note::
            - This is a pure Python implementation and slower than the .rne()
            which is written in C.
            - This version supports symbolic model parameters
            - Verified against MATLAB code

        :seealso: :func:`rne`
        """

        def removesmall(x):
            return x

        n = self.n

        if self.symbolic:
            dtype = 'O'
        else:
            dtype = None

        z0 = np.array([0, 0, 1], dtype=dtype)

        if gravity is None:
            gravity = self.gravity  # default gravity from the object
        else:
            gravity = getvector(gravity, 3)

        if fext is None:
            fext = np.zeros((6,), dtype=dtype)
        else:
            fext = getvector(fext, 6)

        if debug:
            print('grav', gravity)
            print('fext', fext)

        # unpack the joint coordinates and derivatives
        if Q is not None and QD is None and QDD is None:
            # single argument case
            Q = getmatrix(Q, (None, self.n * 3))
            q = Q[:, 0:n]
            qd = Q[:, n:2 * n]
            qdd = Q[:, 2 * n:]

        else:
            # 3 argument case
            q = getmatrix(Q, (None, self.n))
            qd = getmatrix(QD, (None, self.n))
            qdd = getmatrix(QDD, (None, self.n))

        nk = q.shape[0]

        tau = np.zeros((nk, n), dtype=dtype)
        if basewrench:
            wbase = np.zeros((nk, n), dtype=dtype)

        for k in range(nk):
            # take the k'th row of data
            q_k = q[k, :]
            qd_k = qd[k, :]
            qdd_k = qdd[k, :]

            if debug:
                print('q_k', q_k)
                print('qd_k', qd_k)
                print('qdd_k', qdd_k)
                print()

            # joint vector quantities stored columwise in matrix
            #  m suffix for matrix
            Fm = np.zeros((3, n), dtype=dtype)
            Nm = np.zeros((3, n), dtype=dtype)
            # if robot.issym
            #     pstarm = sym([]);
            # else
            #     pstarm = [];
            pstarm = np.zeros((3, n), dtype=dtype)
            Rm = []

            # rotate base velocity and acceleration into L1 frame
            Rb = t2r(self.base.A).T
            # base has zero angular velocity
            w = Rb @ np.zeros((3,), dtype=dtype)
            # base has zero angular acceleration
            wd = Rb @ np.zeros((3,), dtype=dtype)
            vd = Rb @ gravity

            # ----------------  initialize some variables ----------------- #

            for j in range(n):
                link = self.links[j]

                # compute the link rotation matrix
                if link.sigma == 0:
                    # revolute axis
                    Tj = link.A(q_k[j]).A
                    d = link.d
                else:
                    # prismatic
                    Tj = link.A(link.theta).A
                    d = q_k[j]

                # compute pstar:
                #   O_{j-1} to O_j in {j}, negative inverse of link xform
                alpha = link.alpha
                if self.mdh:
                    pstar = np.r_[
                        link.a, -d * sym.sin(alpha), d * sym.cos(alpha)]
                    if j == 0:
                        if self._base:
                            Tj = self._base.A @ Tj
                            pstar = self._base.A @ pstar
                else:
                    pstar = np.r_[
                        link.a, d * sym.sin(alpha), d * sym.cos(alpha)]

                # stash them for later
                Rm.append(t2r(Tj))
                pstarm[:, j] = pstar

            # -----------------  the forward recursion -------------------- #

            for j, link in enumerate(self.links):

                Rt = Rm[j].T    # transpose!!
                pstar = pstarm[:, j]
                r = link.r

                # statement order is important here

                if self.mdh:
                    if link.isrevolute:
                        # revolute axis
                        w_ = Rt @ w + z0 * qd_k[j]
                        wd_ = Rt @ wd \
                            + z0 * qdd_k[j] \
                            + _cross(Rt @ w, z0 * qd_k[j])
                        vd_ = Rt @ _cross(wd, pstar) \
                            + _cross(w, _cross(w, pstar)) \
                            + vd
                    else:
                        # prismatic axis
                        w_ = Rt @ w
                        wd_ = Rt @ wd
                        vd_ = Rt @ (
                            _cross(wd, pstar)
                            + _cross(w, _cross(w, pstar))
                            + vd
                            ) \
                            + 2 * _cross(Rt @ w, z0 * qd_k[j]) \
                            + z0 * qdd_k[j]
                    # trailing underscore means new value, update here
                    w = w_
                    wd = wd_
                    vd = vd_
                else:
                    if link.isrevolute:
                        # revolute axis
                        wd = Rt @ (
                            wd + z0 * qdd_k[j]
                            + _cross(w, z0 * qd_k[j]))
                        w = Rt @ (w + z0 * qd_k[j])
                        vd = _cross(wd, pstar) \
                            + _cross(w, _cross(w, pstar)) \
                            + Rt @ vd
                    else:
                        # prismatic axis
                        w = Rt @ w
                        wd = Rt @ wd
                        vd = Rt @  (z0 * qdd_k[j] + vd) \
                            + _cross(wd, pstar) \
                            + 2 * _cross(w, Rt @ z0 * qd_k[j]) \
                            + _cross(w, _cross(w, pstar))

                vhat = _cross(wd, r) \
                    + _cross(w, _cross(w, r)) \
                    + vd
                Fm[:, j] = link.m * vhat
                Nm[:, j] = link.I @ wd + _cross(w, link.I @ w)

                if debug:
                    print('w:     ', removesmall(w))
                    print('wd:    ', removesmall(wd))
                    print('vd:    ', removesmall(vd))
                    print('vdbar: ', removesmall(vhat))
                    print()

            if debug:
                print('Fm\n', Fm)
                print('Nm\n', Nm)

            # -----------------  the backward recursion -------------------- #

            f = fext[:3]      # force/moments on end of arm
            nn = fext[3:]

            for j in range(n - 1, -1, -1):
                link = self.links[j]
                r = link.r

                #
                # order of these statements is important, since both
                # nn and f are functions of previous f.
                #
                if self.mdh:
                    if j == (n - 1):
                        R = np.eye(3, dtype=dtype)
                        pstar = np.zeros((3,), dtype=dtype)
                    else:
                        R = Rm[j + 1]
                        pstar = pstarm[:, j + 1]

                    f_ = R @ f + Fm[:, j]
                    nn_ = R @ nn \
                        + _cross(pstar, R @ f) \
                        + _cross(pstar, Fm[:, j]) \
                        + Nm[:, j]
                    f = f_
                    nn = nn_

                else:
                    pstar = pstarm[:, j]
                    if j == (n - 1):
                        R = np.eye(3, dtype=dtype)
                    else:
                        R = Rm[j + 1]

                    nn = R @ (nn + _cross(R.T @ pstar, f)) \
                        + _cross(pstar + r, Fm[:, j]) \
                        + Nm[:, j]
                    f = R @ f + Fm[:, j]

                if debug:
                    print('f: ', removesmall(f))
                    print('n: ', removesmall(nn))

                R = Rm[j]
                if self.mdh:
                    if link.isrevolute:
                        # revolute axis
                        t = nn @ z0
                    else:
                        # prismatic
                        t = f @ z0
                else:
                    if link.isrevolute:
                        # revolute axis
                        t = nn @ (R.T @ z0)
                    else:
                        # prismatic
                        t = f @ (R.T @ z0)

                # add joint inertia and friction
                #  no Coulomb friction if model is symbolic
                tau[k, j] = t \
                    + link.G ** 2 * link.Jm * qdd_k[j] \
                    - link.friction(qd_k[j], coulomb=not self.symbolic)
                if debug:
                    print(
                        f'j={j:}, G={link.G:}, Jm={link.Jm:}, friction={link.friction(qd_k[j], coulomb=False):}')  # noqa
                    print()

            # compute the base wrench and save it
            if basewrench:
                R = Rm[0]
                nn = R @ nn
                f = R @ f
                wbase[k, :] = np.r_[f, nn]

        # if self.symbolic:
        #     # simplify symbolic expressions
        #     print(
        #       'start symbolic simplification, this might take a while...')
        #     # from sympy import trigsimp

        #     # tau = trigsimp(tau)
        #     # consider using multiprocessing to spread over cores
        #     #  https://stackoverflow.com/questions/33844085/using-multiprocessing-with-sympy  # noqa
        #     print('done')
        #     if tau.shape[0] == 1:
        #         return tau.reshape(self.n)
        #     else:
        #         return tau

        if tau.shape[0] == 1:
            return tau.flatten()
        else:
            return tau

# -------------------------------------------------------------------------- #

    def ikine_6s(self, T, config, ikfunc):
        # Undo base and tool transformations, but if they are not
        # set, skip the operation.  Nicer for symbolics
        if self._base is not None:
            T = self.base.inv() * T
        if self._tool is not None:
            T = self.tool.inv() * T

        # q = np.zeros((6,))
        solutions = []

        for k, Tk in enumerate(T):
            # get model specific solution for first 3 joints
            theta = ikfunc(self, Tk, config)

            if theta is None or np.any(np.isnan(theta)):
                solution = iksol(None, False, "")
            else:
                # Solve for the wrist rotation
                # We need to account for some random translations between the
                # first and last 3 joints (d4) and also d6,a6,alpha6 in the
                # final frame.

                # Transform of first 3 joints
                T13 = self.A([0, 2], theta)

                # T = T13 * Tz(d4) * R * Tz(d6) Tx(a5)
                Td4 = SE3(0, 0, self.links[3].d)      # Tz(d4)

                # Tz(d6) Tx(a5) Rx(alpha6)
                Tt = SE3(self.links[5].a, 0, self.links[5].d) * \
                    SE3.Rx(self.links[5].alpha)

                R = Td4.inv() * T13.inv() * Tk * Tt.inv()

                # The spherical wrist implements Euler angles
                if 'f' in config:
                    eul = R.eul(flip=True)
                else:
                    eul = R.eul()
                theta = np.r_[theta, eul]
                if self.links[3].alpha > 0:
                    theta[4] = -theta[4]

                # Remove the link offset angles
                theta = theta - self.offset

                solution = iksol(theta, True, "")
                solutions.append(solution)

        if len(solutions) == 1:
            return solutions[0]
        else:
            return solutions


class SerialLink(DHRobot):
    def __init__(self, *args, **kwargs):
        print('SerialLink is deprecated, use DHRobot instead')
        super().__init__(*args, **kwargs)


def _cross(a, b):
    return np.r_[
        a[1] * b[2] - a[2] * b[1],
        a[2] * b[0] - a[0] * b[2],
        a[0] * b[1] - a[1] * b[0]]


if __name__ == "__main__":   # pragma nocover

    import roboticstoolbox as rtb
    # import spatialmath.base.symbolic as sym

    planar = rtb.models.DH.Planar2()
    # J = puma.jacob0(puma.qn)
    # print(J)
    # print(puma.manipulability(puma.qn))
    # print(puma.manipulability(puma.qn, 'asada'))
    # tw, T0 = puma.twists(puma.qz)
    print(planar)

    puma = rtb.models.DH.Puma560()
    print(puma)
    puma.base = None
    print('base', puma.base)
    print('tool', puma.tool)

    print(puma.ets())

    puma[2].flip = True
    puma[3].offset = 1
    puma[4].flip = True
    puma[4].offset = -1
    print(puma)
    print(puma.ets())

    print(puma.dyntable())



#!/usr/bin/env python
"""
@author Jesse Haviland
"""

import numpy as np
import roboticstoolbox as rp
from spatialmath import SE3


class RobotPlot(object):

    def __init__(
            self, robot, ax, readonly, display=True,
            jointaxes=True, eeframe=True, shadow=True, name=True):

        super(RobotPlot, self).__init__()

        # Readonly - True for this robot is for displaying only
        self.readonly = readonly

        # To show to robot in the plot or not
        # If not displayed, the robot is still simulated
        self.display = display

        self.robot = robot
        self.ax = ax

        # Line plot of robot links
        self.links = None

        # Z-axis Coordinate frame (quiver) of joints
        self.joints = []

        # Text of the robots name
        self.name = None

        # Shadow of the the line plot on the x-y axis
        self.sh_links = None

        # Coordinate frame of the ee (three quivers)
        self.ee_axes = []

        # Robot has been drawn
        self.drawn = False

        # Display options
        self.eeframe = eeframe
        self.jointaxes = jointaxes
        self.shadow = shadow
        self.showname = name

    def axes_calcs(self):
        # Joint and ee poses
        T = self.robot.fkine_all(self.robot.q)

        try:
            Te = self.robot.fkine(self.robot.q)
        except ValueError:
            print(
                "\nError: Branched robot's not yet supported "
                "with PyPlot backend\n")
            raise

        Tb = self.robot.base

        # Joint and ee position matrix
        loc = np.zeros([3, len(self.robot.links) + 1])
        loc[:, 0] = Tb.t

        # Joint axes position matrix
        joints = np.zeros((3, self.robot.n))

        # Axes arrow transforms
        Tjx = SE3.Tx(0.06)
        Tjy = SE3.Ty(0.06)
        Tjz = SE3.Tz(0.06)

        # ee axes arrows
        Tex = Te * Tjx
        Tey = Te * Tjy
        Tez = Te * Tjz

        # Joint axes arrow calcs
        if isinstance(self.robot, rp.ERobot):
            i = 0
            j = 0
            for link in self.robot.links:
                loc[:, i + 1] = link._fk.t

                if link.isjoint:
                    if link.v.axis == 'Rz' or link.v.axis == 'tz':
                        Tji = link._fk * Tjz

                    elif link.v.axis == 'Ry' or link.v.axis == 'ty':
                        Tji = link._fk * Tjy

                    elif link.v.axis == 'Rx' or link.v.axis == 'tx':
                        Tji = link._fk * Tjx

                    joints[:, j] = Tji.t
                    j += 1

                i += 1
            loc = np.c_[loc, loc[:, -1]]
        else:
            # End effector offset (tool of robot)
            loc = np.c_[loc, Te.t]

            for i in range(self.robot.n):
                loc[:, i + 1] = T[i].t
                Tji = T[i] * Tjz
                joints[:, i] = Tji.t

        return loc, joints, [Tex, Tey, Tez]

    def draw(self):
        if not self.display:
            return

        if not self.drawn:
            self.init()
            return

        loc, joints, ee = self.axes_calcs()

        # Remove old ee coordinate frame
        if self.eeframe:
            self.ee_axes[0].remove()
            self.ee_axes[1].remove()
            self.ee_axes[2].remove()

            # Plot ee coordinate frame
            self.ee_axes[0] = \
                self._plot_quiver(
                    loc[:, -1], ee[0].t, '#EE9494', 2)
            self.ee_axes[1] = \
                self._plot_quiver(
                    loc[:, -1], ee[1].t, '#93E7B0', 2)
            self.ee_axes[2] = \
                self._plot_quiver(
                    loc[:, -1], ee[2].t, '#54AEFF', 2)

        # Remove oldjoint z coordinates
        if self.jointaxes:
            j = 0

            for joint in self.joints:
                self.ax.collections.remove(joint)

            del self.joints
            self.joints = []

            # Plot joint z coordinates
            for i in range(len(self.robot.links)):
                if isinstance(self.robot, rp.DHRobot) or \
                        self.robot.links[i].isjoint:
                    self.joints.append(
                        self._plot_quiver(
                            loc[:, i+1], joints[:, j], '#8FC1E2', 2))
                    j += 1

            # for i in range(len(self.robot.links)):
            #     self.joints[i] = \
            #         self._plot_quiver(loc[:, i+1], joints[:, i], '#8FC1E2', 2)

        # Update the robot links
        self.links[0].set_xdata(loc[0, :])
        self.links[0].set_ydata(loc[1, :])
        self.links[0].set_3d_properties(loc[2, :])

        # Update the shadow of the robot links
        if self.shadow:
            self.sh_links[0].set_xdata(loc[0, :])
            self.sh_links[0].set_ydata(loc[1, :])
            self.sh_links[0].set_3d_properties(0)

    def draw2(self):
        if not self.display:
            return

        if not self.drawn:
            self.init2()
            return

        loc, joints, ee = self.axes_calcs()

        # Remove old ee coordinate frame
        if self.eeframe:
            self.ee_axes[0].remove()
            self.ee_axes[1].remove()

            # Plot ee coordinate frame
            self.ee_axes[0] = \
                self._plot_quiver2(
                    loc[:, -1], ee[0].t, '#EE9494', 2)
            self.ee_axes[1] = \
                self._plot_quiver2(
                    loc[:, -1], ee[1].t, '#93E7B0', 2)

        # Update the robot links
        self.links[0].set_xdata(loc[0, :])
        self.links[0].set_ydata(loc[1, :])

    def init(self):

        self.drawn = True

        # Joint and ee poses
        Tb = self.robot.base
        loc, joints, ee = self.axes_calcs()

        # Plot robot name
        if self.showname:
            self.name = self.ax.text(
                Tb.t[0], Tb.t[1], 0.05, self.robot.name)

        # Plot ee coordinate frame
        if self.eeframe:
            self.ee_axes.append(
                self._plot_quiver(
                    loc[:, -1], ee[0].t, '#EE9494', 2))
            self.ee_axes.append(
                self._plot_quiver(
                    loc[:, -1], ee[1].t, '#93E7B0', 2))
            self.ee_axes.append(
                self._plot_quiver(
                    loc[:, -1], ee[2].t, '#54AEFF', 2))

        # Plot joint z coordinates
        if self.jointaxes:
            j = 0
            for i in range(len(self.robot.links)):
                if isinstance(self.robot, rp.DHRobot) or \
                        self.robot.links[i].isjoint:
                    self.joints.append(
                        self._plot_quiver(
                            loc[:, i+1], joints[:, j], '#8FC1E2', 2))
                    j += 1

        # Plot the shadow of the robot links, draw first so robot is always
        # in front
        if self.shadow:
            self.sh_links = self.ax.plot(
                loc[0, :], loc[1, :],
                linewidth=3, color='lightgrey')

        # Plot the robot links
        self.links = self.ax.plot(
            loc[0, :], loc[1, :], loc[2, :], linewidth=5, color='#E16F6D')

    def init2(self):

        self.drawn = True

        # Joint and ee poses
        Tb = self.robot.base
        loc, joints, ee = self.axes_calcs()

        # Plot robot name
        if self.showname:
            self.name = self.ax.text(
                Tb.t[0] + 0.05, Tb.t[1], self.robot.name)

        # Plot ee coordinate frame
        if self.eeframe:
            self.ee_axes.append(
                self._plot_quiver2(
                    loc[:, -1], ee[0].t, '#EE9494', 2))
            self.ee_axes.append(
                self._plot_quiver2(
                    loc[:, -1], ee[1].t, '#93E7B0', 2))

        # Plot the robot links
        self.links = self.ax.plot(
            loc[0, :], loc[1, :], linewidth=5, color='#E16F6D')

    def _plot_quiver(self, p0, p1, col, width):
        qv = self.ax.quiver(
            p0[0], p0[1], p0[2],
            p1[0] - p0[0],
            p1[1] - p0[1],
            p1[2] - p0[2],
            linewidth=width,
            color=col
        )

        return qv

    def _plot_quiver2(self, p0, p1, col, width):
        qv = self.ax.quiver(
            p0[0], p0[1],
            p1[0] - p0[0],
            p1[1] - p0[1],
            linewidth=width,
            color=col
        )

        return qv



"""
Python Navigation Abstract Class
@Author: Kristian Gibson
TODO: Comments + Sphynx Docs Structured Text
TODO: Bug-fix, testing

Not ready for use yet.
"""
from scipy import integrate
from spatialmath.base.transforms2d import *
from spatialmath.base.vectors import *
from spatialmath.pose2d import SE2
from scipy.ndimage import binary_dilation
from matplotlib import cm
from abc import ABCMeta, abstractmethod
import matplotlib as mpl
import copy

class Navigation:
    # __metaclass__ = ABCMeta

    def __init__(self, occ_grid=None, goal=None, inflate=0,
                 reset=False, verbose=None, seed=None,
                 transform=SE2()):
        self._occ_grid_nav = None
        self._start = None
        self._verbose = verbose
        self._seed = seed
        self._spin_count = None
        self._rand_stream = None
        self._seed_0 = None
        self._w2g = None
        self._inflate = inflate
        self._reset = reset
        self._transform = transform
        self._goal = None

        # if occ_grid is not None:
        #     # this is the inverse of the matlab code
        #     if type(occ_grid) is dict:
        #         self._occ_grid = occ_grid["map"]
        #         # self._w2g = self._T  # TODO
        #     else:
        #         self._occ_grid = occ_grid
        #         # self._w2g = SE2(0, 0, 0)
        self._occ_grid = occ_grid

        if inflate > 0:
            # Generate a circular structuring element
            r = inflate
            y, x = np.ogrid[-r: r+1, -r: r+1]
            SE = np.square(x) + np.square(y) <= np.square(r)
            SE = SE.astype(int)

            # do the inflation using SciPy
            self._occ_grid_nav = binary_dilation(self._occ_grid, SE)
        else:
            self._occ_grid_nav = self._occ_grid

        if goal is not None:
            self._goal = np.transpose(goal)

        # Simplification of matlab code
        self._privaterandom = np.random.default_rng(seed=seed)

        rs = np.random.RandomState()
        if seed is not None:
            rs = np.random.RandomState(seed)

        self._seed_0 = rs.get_state()
        self._rand_stream = rs
        self._w2g = transform
        self._spin_count = 0

    def __str__(self):
        s = f"Navigation class: {self.__class__.__name__}"
        if self._occ_grid is not None:
            s += f"\n  Occupancy grid: {self._occ_grid.shape}"
        if self._goal is not None:
            s += f"\n  Goal: {self.goal}"
        return s

    @property
    def occ_grid(self):
        return self._occ_grid

    @property
    def occ_grid_nav(self):
        return self._occ_grid_nav

    @property
    def goal(self):
        return self._goal

    @goal.setter
    def goal(self, goal):
        if self.is_occupied(goal):
            raise ValueError("Goal location inside obstacle")
        self._goal = base.getvector(goal, 2)

    def select_goal(self):
        self.plot()
        print("Select goal location")
        goal = round(plt.ginput(1))

        self.goal = goal

    @property
    def start(self):
        return self._start

    @property
    def verbose(self):
        return self._verbose

    @property
    def seed(self):
        return self._seed

    @property
    def spin_count(self):
        return self._spin_count

    @property
    def rand_stream(self):
        return self._rand_stream

    @property
    def seed_0(self):
        return self._seed_0

    @property
    def w2g(self):
        return self._w2g

    @property
    def inflate(self):
        return self._inflate

    @property
    def private(self):
        return self._private

    @property
    def reset(self):
        return self._reset

    @property
    def transform(self):
        return self._transform

    # Define abstract classes to be implemented later
    @abstractmethod
    def plan(self):
        pass

    @abstractmethod
    def next(self):
        pass

    def query(self, start, goal=None, animate=False):
        """
        ssss

        :param start: [description]
        :type start: [type]
        :param goal: [description], defaults to None
        :type goal: [type], optional
        :param animate: [description], defaults to False
        :type animate: bool, optional
        :return: [description]
        :rtype: [type]

                    %Navigation.query Find a path from start to goal using plan
            %
            % N.query(START, OPTIONS) animates the robot moving from START (2x1) to the goal (which is a 
            % property of the object) using a previously computed plan.
            %
            % X = N.query(START, OPTIONS) returns the path (Mx2) from START to the goal (which is a property of 
            % the object).
            %
            % The method performs the following steps:
            %  - Initialize navigation, invoke method N.navigate_init()
            %  - Visualize the environment, invoke method N.plot()
            %  - Iterate on the next() method of the subclass until the goal is
            %    achieved.
            %
            % Options::
            % 'animate'    Show the computed path as a series of green dots.
            %
            % Notes::
            %  - If START given as [] then the user is prompted to click a point on the map.
            %
            %
            % See also Navigation.navigate_init, Navigation.plot, Navigation.goal.
        """
        if goal is not None:
            self.check_query(start, goal)
        else:
            self.check_query(start, self.goal)

        if animate:
            self.plot()

        robot = self._start
        path = [robot]

        while True:
            if animate:
                plt.plot(robot[0], robot[1], 'y.', 12)
                plt.pause(0.05)

            # get next point on the path
            robot = self.next(robot)

            # are we are done?
            if robot is None:
                path.append(self._goal)
                return np.array(path)

            path.append(robot)

    def plot(self, block=False, **kwargs):
        self.plot_bg(**kwargs)
        self.plot_fg(**kwargs)
        plt.show(block=block)


    def plot_bg(self, distance=None, color_map='gray',
                inflated=True, path_marker=None, **unused_args):
        """
        [summary]

        :param p: [description], defaults to None
        :type p: [type], optional
        :param distance: [description], defaults to None
        :type distance: [type], optional
        :param color_map: [description], defaults to cm.get_cmap('bone')
        :type color_map: [type], optional
        :param beta: [description], defaults to 0.2
        :type beta: float, optional
        :param inflated: [description], defaults to False
        :type inflated: bool, optional

                %Navigation.plot  Visualization background
        %
        % N.plot_bg(OPTIONS) displays the occupancy grid with occupied cells shown as
        % red and an optional distance field.
        %
        % N.plot_bg(P,OPTIONS) as above but overlays the points along the path (2xM) matrix. 
        %
        % Options::
        %  'distance',D      Display a distance field D behind the obstacle map.  D is
        %                    a matrix of the same size as the occupancy grid.
        %  'colormap',@f     Specify a colormap for the distance field as a function handle, eg. @hsv
        %  'beta',B          Brighten the distance field by factor B.
        %  'inflated'        Show the inflated occupancy grid rather than original
        %  'pathmarker',M    Options to draw a path point
        %  'startmarker',M   Options to draw the start marker
        %  'goalmarker',M    Options to draw the goal marker
        %
        % Notes::
        % - The distance field at a point encodes its distance from the goal, small
        %   distance is dark, a large distance is bright.  Obstacles are encoded as
        %   red.
        % - Beta value -1<B<0 to darken, 0<B<+1 to lighten.
        %
        % See also Navigation.plot, Navigation.plot_fg, brighten.
        """

        fig, ax = plt.subplots(nrows=1, ncols=1)

        if distance is None:
            # occupancy grid only

            # create color map for free space + obstacle:
            #   free space, color index = 1, white, alpha=0 to allow background and grid lines to show
            #   obstacle, color index = 2, red, alpha=1

            if self._inflate and inflated:
                # 0 background (white, transparent)
                # 1 inflated obstacle (pink)
                # 2 original obstacle (red)
                colors = [(1, 1, 1, 0), (1, 0.75, 0.8, 1), (1, 0, 0, 1)]
                image = (self._occ_grid + self._occ_grid_nav)
            else:
                # 0 background
                # 1 obstacle
                colors = [(1, 1, 1, 0), (1, 0, 0, 1)]
                image = self._occ_grid

            ax.set_facecolor((1, 1, 1)) # create white background
            c_map = mpl.colors.ListedColormap(colors)
            ax.imshow(image, cmap=c_map, origin='lower', zorder=1)

            ax.grid(True, zorder=0)

        else:
            # distance field with obstacles

            # find largest finite value
            v = distance.flatten()
            vmax = max(v[np.isfinite(v)])

            # create a copy of greyscale color map
            c_map = copy.copy(mpl.cm.get_cmap(color_map))
            c_map.set_bad(color=(1,0,0,1))  # nan and inf are red

            # change all inf to large value, so they are not 'bad' ie. red
            distance[np.isinf(distance)] = 2 * vmax
            c_map.set_over(color=(0,0,1))  # ex-infs are now blue

            # display image
            ax.imshow(distance, origin='lower',
                cmap=c_map,
                norm=mpl.colors.Normalize(vmin=0, vmax=vmax, clip=False)
                )
            ax.grid(True, alpha=0.1, color=(1,1,1))

            # add colorbar
            scalar_mappable_c_map = cm.ScalarMappable(cmap=c_map)
            plt.colorbar(scalar_mappable_c_map)

        ax.set_xlabel('x (cells)')
        ax.set_ylabel('y (cells)')

        # lock axis limits to current value
        ax.set_xlim(ax.get_xlim())
        ax.set_ylim(ax.get_ylim())

        plt.draw()
        plt.show(block=False)


    def plot_fg(self, path=None, path_marker=None, start_marker=None, goal_marker=None, **unused_args):
        """
        [summary]

        :param p: [description], defaults to None
        :type p: [type], optional
        :param path_marker: [description], defaults to None
        :type path_marker: [type], optional
        :param start_marker: [description], defaults to None
        :type start_marker: [type], optional
        :param goal_marker: [description], defaults to None
        :type goal_marker: [type], optional
        :param goal: [description], defaults to None
        :type goal: [type], optional

                %Navigation.plot_fg  Visualization foreground
        %
        % N.plot_fg(OPTIONS) displays the start and goal locations if specified.
        % By default the goal is a pentagram and start is a circle.
        %
        % N.plot_fg(P, OPTIONS) as above but overlays the points along the path (2xM) matrix.
        % 
        % Options::
        %  'pathmarker',M    Options to draw a path point
        %  'startmarker',M   Options to draw the start marker
        %  'goalmarker',M    Options to draw the goal marker
        %
        % Notes::
        % - In all cases M is a single string eg. 'r*' or a cell array of MATLAB LineSpec options.
        % - Typically used after a call to plot_bg().
        %
        % See also Navigation.plot_bg.
        """

        # create default markers
        if path_marker is None:
            path_marker = { 'marker': '.',
                            'markerfacecolor': 'y',
                            'markersize': 12,
                          }
        if start_marker is None:
            start_marker = {'marker': 'o',
                            'markeredgecolor': 'w',
                            'markerfacecolor': 'b', 
                            'markersize': 10,
                           }
        if goal_marker is None:
            goal_marker = { 'marker': '*',
                            'markeredgecolor': 'w',
                            'markerfacecolor': 'b',
                            'markersize': 16,
                          }

        # mark the path
        if path is not None:
            if not isinstance(path, np.ndarray) or not (path.shape[1] in (2,3)):
                raise ValueError('path must be an Nx2 or Nx3 matrix of points')
            if path.shape[1] == 2:
                plt.plot(path[:, 0], path[:, 1], **path_marker)
            elif path.shape[1] == 3:
                plt.plot3(path[:, 0], path[:, 1], path[:, 2], **path_marker)

        # mark start and goal if requested
        if self._goal is not None:
            if len(self._goal) == 2:
                plt.plot(self._goal[0], self._goal[1], **goal_marker)
            else:
                plt.plot3(self._goal[0], self._goal[1], self._goal[2]+0.1, **goal_marker)
        if self._start is not None:
            if len(self._start) == 2:
                plt.plot(self._start[0], self._start[1], **start_marker)
            else:
                plt.plot3(self._start[0], self._start[1], self._goal[2]+0.1, **start_marker)

        plt.draw()


    def check_query(self, start=None, goal=None):
        # if start is None:
        #     self.plot()
        #     disp("Select start location")
        #     start = round(plt.ginput(1))

        # if goal is None:
        #     self.plot()
        #     disp("Select goal location")
        #     goal = round(plt.ginput(1))
        # TODO

        self._start = base.getvector(start, 2, dtype=np.int)
        self._goal = base.getvector(goal, 2, dtype=np.int)

        if self.is_occupied(self._start):
               raise ValueError("Start location inside obstacle")
        if self.is_occupied(self.goal):
            raise ValueError("Goal location inside obstacle")

    def is_occupied(self, p):
        """
        [summary]

        :param p: [description]
        :type p: [type]
        :return: [description]
        :rtype: [type]

                    %Navigation.isoccupied Test if grid cell is occupied
            %
            % N.isoccupied(POS) is true if there is a valid grid map and the
            % coordinates given by the columns of POS (2xN) are occupied.
            %
            % N.isoccupied(X,Y) as above but the coordinates given separately.
            %
            % Notes:
            % -  X and Y are Cartesian rather than MATLAB row-column coordinates.
        """
        occ = None
        pis = None
        if self._occ_grid_nav is None:
            return False

        p = base.getvector(p, 2, 'list')
        try:
            return self._occ_grid_nav[p[1], p[0]] > 0
        except IndexError:
            return True  # points outside the map are considered occupied

        # if x is not None:
        #     if np.size(x) == 2:
        #         x = np.transpose(x)
        #     assert(np.shape(x, 0) == 2, "RTB:Navigation:isoccupied. P must have 2 rows")
        #     pos = x
        # else:
        #     assert(np.size(x) == np.size(y), "RTB:Navigation:isoccupied. X and Y must be same length")
        #     pos = np.array([(np.transpose(x)), np.transpose(y)])

        # pos = round(self._w2g * pos)
        # k = pos[0, :] > 0 & pos[0, :] <= np.shape(self._occ_grid) & pos[1,:] > 0 <= np.shape(self._occ_grid, 1)

        # i = sub2ind(np.shape(self._occ_grid), pos[1, k], pos[0, k])
        # occ = np.ones(1, np.size(pos, 1))  # TODO: this bit normally says 'logic' in matlab... should be fine
        # occ[k] = self._occ_grid_nav[i] > 0
        # return occ

    def goal_change(self):
        pass

    def navigate_init(self, start):
        pass

    def rand(self, **kwargs):
        return self._privaterandom.uniform(**kwargs)

    def randn(self, **kwargs):
        return self._privaterandom.normal(**kwargs)

    def randi(self, *pos, **kwargs):
        return self._privaterandom.integers(*pos, **kwargs)

    def verbosity(self, v):
        self._verbose = v

    def message(self, s, args=None):
        if self._verbose:
            if args is None:
                print("NavigationDebug:: " + s)
            else:
                print("NavigationDebug:: " + s.format(args))

    def spinner(self):
        spin_chars = "\|/"  # TODO: This might break?
        self._spin_count = self._spin_count + 1
        print(spin_chars[np.mod(self._spin_count, len(spin_chars))+1])

    @staticmethod
    def show_distance(d):
        d[np.isinf(d)] = None
        ax = plt.gca()
        c_map = plt.get_cmap("Greys")
        plt.clim(0, np.max(d[:]))
        plt.figimage(d)
        plt.xlabel('X')
        plt.ylabel('Y')
        plt.show()

    # There's no waitbar in matplotlib, so just going with placeholder functions
    @staticmethod
    def progress_init(title):
        h = "Waiting"
        return h

    @staticmethod
    def progress(h, x):
        pass

    @staticmethod
    def progress_delete(h):
        plt.clf()



# Sourced from: https://stackoverflow.com/questions/28995146/matlab-ind2sub-equivalent-in-python/28995315#28995315
def sub2ind(array_shape, rows, cols):
    ind = rows*array_shape[1] + cols
    ind[ind < 0] = -1
    ind[ind >= array_shape[0]*array_shape[1]] = -1
    return ind

class Error(Exception):
    """Base class for other exceptions"""
    pass




#!/usr/bin/env python

import numpy as np
from roboticstoolbox.robot.ERobot import ERobot


class Mico(ERobot):
    """
    Class that imports a Mico URDF model

    ``Panda()`` is a class which imports a Kinova Mico robot definition
    from a URDF file.  The model describes its kinematic and graphical
    characteristics.

    .. runblock:: pycon

        >>> import roboticstoolbox as rtb
        >>> robot = rtb.models.URDF.Mico()
        >>> print(robot)

    Defined joint configurations are:

    - qz, zero joint angle configuration, 'L' shaped configuration
    - qr, vertical 'READY' configuration

    .. codeauthor:: Jesse Haviland
    .. sectionauthor:: Peter Corke
    """
    def __init__(self):

        args = super().urdf_to_ets_args(
            "kinova_description/urdf/j2n4s300_standalone.xacro")

        super().__init__(
                args[0],
                name=args[1],
                manufacturer='Kinova'
            )

        self.addconfiguration(
            "qz", np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0]))
        self.addconfiguration(
            "qr", np.array([0, 45, 60, 0, 0, 0, 0, 0, 0, 0]) * np.pi/180)


if __name__ == '__main__':   # pragma nocover

    robot = Mico()
    print(robot)



#!/usr/bin/env python

import numpy as np
from spatialmath.base import trotz, transl
from roboticstoolbox import DHRobot, RevoluteMDH


class Panda(DHRobot):
    """
    A class representing the Panda robot arm.

    ``Panda()`` is a class which models a Franka-Emika Panda robot and
    describes its kinematic characteristics using modified DH
    conventions.

    .. runblock:: pycon

        >>> import roboticstoolbox as rtb
        >>> robot = rtb.models.DH.Panda()
        >>> print(robot)

    .. note::
        - SI units of metres are used.
        - The model includes a tool offset.

    :references:
        - https://frankaemika.github.io/docs/control_parameters.html

    .. codeauthor:: Samuel Drew
    .. codeauthor:: Peter Corke
    """

    def __init__(self):

        # deg = np.pi/180
        mm = 1e-3
        tool_offset = (103)*mm

        flange = (107)*mm
        # d7 = (58.4)*mm

        # This Panda model is defined using modified
        # Denavit-Hartenberg parameters
        L = [
                RevoluteMDH(
                    a=0.0,
                    d=0.333,
                    alpha=0.0,
                    qlim=np.array([-2.8973, 2.8973])
                ),

                RevoluteMDH(
                    a=0.0,
                    d=0.0,
                    alpha=-np.pi/2,
                    qlim=np.array([-1.7628, 1.7628])
                ),

                RevoluteMDH(
                    a=0.0,
                    d=0.316,
                    alpha=np.pi/2,
                    qlim=np.array([-2.8973, 2.8973])
                 ),

                RevoluteMDH(
                    a=0.0825,
                    d=0.0,
                    alpha=np.pi/2,
                    qlim=np.array([-3.0718, -0.0698])
                ),

                RevoluteMDH(
                    a=-0.0825,
                    d=0.384,
                    alpha=-np.pi/2,
                    qlim=np.array([-2.8973, 2.8973])
                ),

                RevoluteMDH(
                    a=0.0,
                    d=0.0,
                    alpha=np.pi/2,
                    qlim=np.array([-0.0175, 3.7525])
                ),

                RevoluteMDH(
                    a=0.088,
                    d=flange,
                    alpha=np.pi/2,
                    qlim=np.array([-2.8973, 2.8973])
                )
        ]

        tool = transl(0, 0, tool_offset) @  trotz(-np.pi/4)

        super().__init__(
            L,
            name='Panda',
            manufacturer='Franka Emika',
            meshdir='meshes/FRANKA-EMIKA/Panda',
            tool=tool)

        self.addconfiguration("qz", [0, 0, 0, 0, 0, 0, 0])
        self.addconfiguration("qr", np.r_[0, -0.3, 0, -2.2, 0, 2.0, np.pi/4])


if __name__ == '__main__':   # pragma nocover

    panda = Panda()
    print(panda)




from roboticstoolbox.robot.Robot import Robot
from ansitable import ANSITable, Column
# import importlib


def list(keywords=None, dof=None, mtype=None):
    """
    Display all robot models in summary form

    :param keywords: keywords to filter on, defaults to None
    :type keywords: tuple of str, optional
    :param dof: number of DoF to filter on, defaults to None
    :type dof: int, optional

    - ``list()`` displays a list of all models provided by the Toolbox.  It
      lists the name, manufacturer, model type, number of DoF, and keywords.

    - ``list(mtype=MT)`` as above, but only displays models of type ``MT``
      where ``MT`` is one of "DH", "ETS" or "URDF".

    - ``list(keywords=KW)`` as above, but only displays models that have a
      keyword in the tuple ``KW``.

    - ``list(dof=N)`` as above, but only display models that have ``N``
      degrees of freedom.

    The filters can be combined

    - ``list(keywords=KW, dof=N)`` are those models that have a keyword in
      ``KW`` and have ``N`` degrees of freedom.
    """

    import roboticstoolbox.models as m
    # module = importlib.import_module(
    #   '.' + os.path.splitext(file)[0], package='bdsim.blocks')

    def make_table(border):
        table = ANSITable(
            Column("class", headalign="^", colalign="<"),
            Column("model", headalign="^", colalign="<"),
            Column("manufacturer", headalign="^", colalign="<"),
            Column("model type", headalign="^", colalign="<"),
            Column("DoF", colalign="<"),
            Column("config", colalign="<"),
            Column("keywords", headalign="^", colalign="<"),
            border=border
        )

        if mtype is not None:
            categories = [mtype]
        else:
            categories = ['DH', 'URDF', 'ETS']
        for category in categories:
            group = m.__dict__[category]
            for cls in group.__dict__.values():
                if isinstance(cls, type) and issubclass(cls, Robot):
                    # we found a Robot subclass, instantiate it
                    robot = cls()
                    try:
                        config = robot.config()
                    except Exception:   # pragma nocover
                        config = ""

                    # apply filters
                    if keywords is not None:
                        if len(set(keywords) & set(robot.keywords)) == 0:
                            continue
                    if dof is not None and robot.n != dof:
                        continue     # pragma nocover

                    # add the row
                    table.row(
                        cls.__name__,
                        robot.name,
                        robot.manufacturer,
                        category,
                        robot.n,
                        config,
                        ', '.join(robot.keywords)
                    )

        print(str(table))

    try:
        make_table('thin')
    except UnicodeEncodeError:
        make_table('ascii')


if __name__ == "__main__":   # pragma nocover
    list()
    list(keywords=('dynamics',))
    list(dof=6)
    list(keywords=('dynamics',), dof=6)



#!/usr/bin/env python
"""
@author Micah Huth
"""


from vpython import color, label, mag, vector
from numpy import sign, arange


def get_text_size(scene):  # pragma nocover
    """
    Determine the text size based on zoom distance

    :param scene: The scene in which the camera to be used resides.
    :type scene: `vpython.canvas`
    """
    # Distance of camera from focus point to determine text size
    distance_from_center = mag(scene.center - scene.camera.pos)

    # Eq generated from data (Using 3rd order polynomial)
    #  D  | size
    #  0  |   5
    # 0.5 |   8
    #  1  |  10
    #  2  |  12
    #  3  |  14
    #  5  |  15
    val = 0.1114 * distance_from_center**3 - 1.336 * distance_from_center**2 \
        + 5.8666 * distance_from_center + 5.1711

    return min(max(val, 10), 15)  # Return val between 10 and 15


def draw_label(label_text, label_position, scene):  # pragma nocover
    """
    Display a label at a given position, with borders and lines

    :param label_text: String of text to be written on the label.
    :type label_text: `str`
    :param label_position: 3D vector position to draw the label at.
    :type label_position: class:`vpython.vector`
    :param scene: The scene in which to draw the object
    :type scene: class:`vpython.canvas`
    :return: The created label object.
    :rtype: class:`vpython.label`
    """

    # Custom settings for the label
    label_height = 10
    label_xoffset = 0
    label_yoffset = 50
    label_space = 20
    label_font = 'serif'
    label_text_colour = color.black
    label_line_color = color.black

    the_label = label(
        canvas=scene,
        pos=label_position,
        text=label_text,
        height=label_height,
        xoffset=label_xoffset,
        yoffset=label_yoffset,
        space=label_space,
        font=label_font,
        color=label_text_colour,
        linecolor=label_line_color
    )

    return the_label


def draw_text(label_text, label_position, scene):  # pragma nocover
    """
    Display a label at a given position, without borders or lines.

    :param label_text: String of text to be written on the label.
    :type label_text: `str`
    :param label_position: 3D vector position to draw the label at.
    :type label_position: class:`vpython.vector`
    :param scene: The scene in which to draw the object
    :type scene: class:`vpython.canvas`
    :return: The created label object.
    :rtype: class:`vpython.label`
    """
    label_height = get_text_size(scene)
    label_xoffset = 0
    label_yoffset = 0
    label_space = 0
    label_font = 'serif'
    label_text_colour = color.black
    label_line_color = color.white
    label_bg_opacity = 0
    label_linewidth = 0.1

    the_label = label(
        canvas=scene,
        pos=label_position,
        text=label_text,
        height=label_height,
        xoffset=label_xoffset,
        yoffset=label_yoffset,
        space=label_space,
        font=label_font,
        color=label_text_colour,
        linecolor=label_line_color,
        opacity=label_bg_opacity,
        linewidth=label_linewidth
    )

    return the_label


def update_grid_numbers(
        focal_point, numbers_list, num_squares,
        scale, is_3d, scene):  # pragma nocover
    """
    Draw the grid numbers along the xyz axes.

    :param focal_point: The focus point of the camera to draw the grid about
    :type focal_point: `list`
    :param numbers_list: A reference to a list of the labels that gets updated.
    :type numbers_list: `list`
    :param num_squares: How many unit squares to draw along the axis.
    :type num_squares: `int`
    :param scale: The scaled length of 1 square unit
    :type scale: `float`
    :param is_3d: Whether the grid is 3D or not
    :type is_3d: `bool`
    :param scene: The scene in which to draw the object
    :type scene: class:`vpython.canvas`
    """

    # Initial conditions
    padding = 0.25  # Padding to not draw numbers on top of lines.
    camera_axes = scene.camera.axis
    # Locate center of the axes
    x_origin, y_origin, z_origin = focal_point[0], focal_point[1], \
        focal_point[2]

    #   CAMERA AXES |  DISPLAYED GRID | XZ PLANE | XY PLANE | YZ PLANE
    #      x,y,z    |      x,y,z      |   x,z    |    x,y   |    y,z
    #  -------------+-----------------+----------+----------+----------
    #      -,-,-    |      +,+,+      |   +,+    |    +,+   |    +,+
    #      -,-,+    |      +,+,-      |   +,-    |    +,+   |    +,-
    #      -,+,-    |      +,-,+      |   +,+    |    +,-   |    -,+
    #      -,+,+    |      +,-,-      |   +,-    |    +,-   |    -,-
    #      +,-,-    |      -,+,+      |   -,+    |    -,+   |    +,+
    #      +,-,+    |      -,+,-      |   -,-    |    -,+   |    +,-
    #      +,+,-    |      -,-,+      |   -,+    |    -,-   |    -,+
    #      +,+,+    |      -,-,-      |   -,-    |    -,-   |    -,-
    # min = -num_squares or 0, around the default position
    # max = +num_squares or 0, around the default position
    # e.g. at the origin, for negative axes: -10 -> 0, positive axes: 0 -> 10
    min_x_coord = round(x_origin + (-(num_squares / 2) +
                                    (sign(camera_axes.x) * -1)
                                    * (num_squares / 2)) * scale, 2)
    max_x_coord = round(x_origin + ((num_squares / 2) +
                                    (sign(camera_axes.x) * -1)
                                    * (num_squares / 2)) * scale, 2)

    min_y_coord = round(y_origin + (-(num_squares / 2) +
                                    (sign(camera_axes.y) * -1)
                                    * (num_squares / 2)) * scale, 2)
    max_y_coord = round(y_origin + ((num_squares / 2) +
                                    (sign(camera_axes.y) * -1)
                                    * (num_squares / 2)) * scale, 2)

    min_z_coord = round(z_origin + (-(num_squares / 2) +
                                    (sign(camera_axes.z) * -1)
                                    * (num_squares / 2)) * scale, 2)
    max_z_coord = round(z_origin + ((num_squares / 2) +
                                    (sign(camera_axes.z) * -1)
                                    * (num_squares / 2)) * scale, 2)

    x_coords = arange(min_x_coord, max_x_coord + scale, scale)
    y_coords = arange(min_y_coord, max_y_coord + scale, scale)
    z_coords = arange(min_z_coord, max_z_coord + scale, scale)

    # If the grid has given too many objects
    if len(x_coords) > num_squares + 1:
        x_coords = x_coords[0:num_squares + 1]
    if len(y_coords) > num_squares + 1:
        y_coords = y_coords[0:num_squares + 1]
    if len(z_coords) > num_squares + 1:
        z_coords = z_coords[0:num_squares + 1]

    # Compound origins are in the middle of the bounding boxes.
    # Thus new pos will be between max and min.
    x_middle = x_coords.mean()
    y_middle = y_coords.mean()
    z_middle = z_coords.mean()

    # If input is empty, append new, otherwise update current
    append = len(numbers_list) == 0
    # Dimensions don't change between updates, so indexing shall
    # remain the same
    index = 0

    # X plane
    for x_pos in x_coords:
        # Draw the corresponding unit number at each x coordinate
        txt = "{:.2f}".format(x_pos)
        if is_3d:
            if (sign(camera_axes.y) * -1) > 0:
                pos = vector(x_pos, max_y_coord + padding, z_origin)
            else:
                pos = vector(x_pos, min_y_coord - padding, z_origin)
        else:
            pos = vector(x_pos, y_origin - padding, z_origin)
        if append:
            numbers_list.append(draw_text(txt, pos, scene))
            numbers_list[len(numbers_list)-1].height = get_text_size(scene)
        else:
            numbers_list[index].text = txt
            numbers_list[index].pos = pos
            numbers_list[index].height = get_text_size(scene)
            index += 1
    # Draw the axis label at the centre of the axes numbers
    txt = "X"
    if (sign(camera_axes.y) * -1) > 0:
        x = x_middle
        y = max_y_coord + scale * 2
        pos = vector(x, y, z_origin)
    else:
        x = x_middle
        y = min_y_coord - scale * 2
        pos = vector(x, y, z_origin)
    if append:
        numbers_list.append(draw_text(txt, pos, scene))
        numbers_list[len(numbers_list) - 1].height = get_text_size(scene)
    else:
        numbers_list[index].text = txt
        numbers_list[index].pos = pos
        numbers_list[index].height = get_text_size(scene)
        index += 1

    # Y plane
    for y_pos in y_coords:
        # Draw the corresponding unit number at each x coordinate
        txt = "{:.2f}".format(y_pos)
        if is_3d:
            if (sign(camera_axes.x) * -1) > 0:
                pos = vector(max_x_coord + padding, y_pos, z_origin)
            else:
                pos = vector(min_x_coord - padding, y_pos, z_origin)
        else:
            pos = vector(x_origin - padding, y_pos, z_origin)
        if append:
            numbers_list.append(draw_text(txt, pos, scene))
            numbers_list[len(numbers_list) - 1].height = get_text_size(scene)
        else:
            numbers_list[index].text = txt
            numbers_list[index].pos = pos
            numbers_list[index].height = get_text_size(scene)
            index += 1
    # Draw the axis label at the centre of the axes numbers
    txt = "Y"
    if (sign(camera_axes.x) * -1) > 0:
        x = max_x_coord + scale * 2
        y = y_middle
        pos = vector(x, y, z_origin)
    else:
        x = min_x_coord - scale * 2
        y = y_middle
        pos = vector(x, y, z_origin)

    if append:
        numbers_list.append(draw_text(txt, pos, scene))
        numbers_list[len(numbers_list) - 1].height = get_text_size(scene)
    else:
        numbers_list[index].text = txt
        numbers_list[index].pos = pos
        numbers_list[index].height = get_text_size(scene)
        index += 1

    if not is_3d:
        return

    # Z plane
    for z_pos in z_coords:
        # Draw the corresponding unit number at each x coordinate
        txt = "{:.2f}".format(z_pos)
        if (sign(camera_axes.x) * -1) > 0:
            pos = vector(max_x_coord + padding, y_origin, z_pos)
        else:
            pos = vector(min_x_coord - padding, y_origin, z_pos)
        if append:
            numbers_list.append(draw_text(txt, pos, scene))
            numbers_list[len(numbers_list) - 1].height = get_text_size(scene)
        else:
            numbers_list[index].text = txt
            numbers_list[index].pos = pos
            numbers_list[index].height = get_text_size(scene)
            index += 1
    # Draw the axis label at either the positive or negative side away from
    # center
    # If sign = -1, draw off max side, if sign = 0 or 1, draw off negative side
    txt = "Z"
    if (sign(camera_axes.x) * -1) > 0:
        pos = vector(max_x_coord + scale * 2, y_origin, z_middle)
    else:
        pos = vector(min_x_coord - scale * 2, y_origin, z_middle)
    if append:
        numbers_list.append(draw_text(txt, pos, scene))
        numbers_list[len(numbers_list) - 1].height = get_text_size(scene)
    else:
        numbers_list[index].text = txt
        numbers_list[index].pos = pos
        numbers_list[index].height = get_text_size(scene)
        index += 1



#!/usr/bin/env python
"""
@author (Original) Matthew Matl, Github: mmatl
@author (Adapted by) Jesse Haviland
"""

import numpy as np
import roboticstoolbox as rtb
import copy
import os
import xml.etree.ElementTree as ET
import spatialmath as sm
from io import BytesIO
from pathlib import Path

from .utils import (parse_origin, configure_origin)

abspath = Path(rtb.__file__).parent / 'models' / 'URDF' / 'xacro'


class URDFType(object):
    """Abstract base class for all URDF types.
    This has useful class methods for automatic parsing/unparsing
    of XML trees.
    There are three overridable class variables:
    - ``_ATTRIBS`` - This is a dictionary mapping attribute names to a tuple,
      ``(type, required)`` where ``type`` is the Python type for the
      attribute and ``required`` is a boolean stating whether the attribute
      is required to be present.
    - ``_ELEMENTS`` - This is a dictionary mapping element names to a tuple,
      ``(type, required, multiple)`` where ``type`` is the Python type for the
      element, ``required`` is a boolean stating whether the element
      is required to be present, and ``multiple`` is a boolean indicating
      whether multiple elements of this type could be present.
      Elements are child nodes in the XML tree, and their type must be a
      subclass of :class:`.URDFType`.
    - ``_TAG`` - This is a string that represents the XML tag for the node
      containing this type of object.
    """
    _ATTRIBS = {}   # Map from attrib name to (type, required)
    _ELEMENTS = {}  # Map from element name to (type, required, multiple)
    _TAG = ''       # XML tag for this element

    def __init__(self):  # pragma nocover
        pass

    @classmethod
    def _parse_attrib(cls, val_type, val):
        """Parse an XML attribute into a python value.
        Parameters
        ----------
        val_type : :class:`type`
            The type of value to create.
        val : :class:`object`
            The value to parse.
        Returns
        -------
        val : :class:`object`
            The parsed attribute.
        """
        if val_type == np.ndarray:
            val = np.fromstring(val, sep=' ')
        else:
            val = val_type(val)
        return val

    @classmethod
    def _parse_simple_attribs(cls, node):
        """Parse all attributes in the _ATTRIBS array for this class.
        Parameters
        ----------
        node : :class:`lxml.etree.Element`
            The node to parse attributes for.
        Returns
        -------
        kwargs : dict
            Map from attribute name to value. If the attribute is not
            required and is not present, that attribute's name will map to
            ``None``.
        """
        kwargs = {}
        for a in cls._ATTRIBS:
            t, r = cls._ATTRIBS[a]  # t = type, r = required (bool)
            if r:
                try:
                    v = cls._parse_attrib(t, node.attrib[a])
                except Exception:   # pragma nocover
                    raise ValueError(
                        'Missing required attribute {} when parsing an object '
                        'of type {}'.format(a, cls.__name__)
                    )
            else:
                v = None
                if a in node.attrib:
                    v = cls._parse_attrib(t, node.attrib[a])
            kwargs[a] = v
        return kwargs

    @classmethod
    def _parse_simple_elements(cls, node, path):
        """Parse all elements in the _ELEMENTS array from the children of
        this node.
        Parameters
        ----------
        node : :class:`lxml.etree.Element`
            The node to parse children for.
        path : str
            The string path where the XML file is located (used for resolving
            the location of mesh or image files).
        Returns
        -------
        kwargs : dict
            Map from element names to the :class:`URDFType` subclass (or list,
            if ``multiple`` was set) created for that element.
        """
        kwargs = {}
        for a in cls._ELEMENTS:
            t, r, m = cls._ELEMENTS[a]
            if not m:
                v = node.find(t._TAG)
                if r or v is not None:
                    v = t._from_xml(v, path)
            else:
                vs = node.findall(t._TAG)
                if len(vs) == 0 and r:   # pragma nocover
                    raise ValueError(
                        'Missing required subelement(s) of type {} when '
                        'parsing an object of type {}'.format(
                            t.__name__, cls.__name__
                        )
                    )
                v = [t._from_xml(n, path) for n in vs]
            kwargs[a] = v
        return kwargs

    @classmethod
    def _parse(cls, node, path):
        """Parse all elements and attributes in the _ELEMENTS and _ATTRIBS
        arrays for a node.
        Parameters
        ----------
        node : :class:`lxml.etree.Element`
            The node to parse.
        path : str
            The string path where the XML file is located (used for resolving
            the location of mesh or image files).
        Returns
        -------
        kwargs : dict
            Map from names to Python classes created from the attributes
            and elements in the class arrays.
        """
        kwargs = cls._parse_simple_attribs(node)
        kwargs.update(cls._parse_simple_elements(node, path))
        return kwargs

    @classmethod
    def _from_xml(cls, node, path):
        """Create an instance of this class from an XML node.
        Parameters
        ----------
        node : :class:`lxml.etree.Element`
            The node to parse.
        path : str
            The string path where the XML file is located (used for resolving
            the location of mesh or image files).
        Returns
        -------
        obj : :class:`URDFType`
            An instance of this class parsed from the node.
        """
        return cls(**cls._parse(node, path))


###############################################################################
# Link types
###############################################################################


class Box(URDFType):
    """A rectangular prism whose center is at the local origin.
    Parameters
    ----------
    size : (3,) float
        The length, width, and height of the box in meters.
    """

    _ATTRIBS = {
        'size': (np.ndarray, True)
    }
    _TAG = 'box'

    def __init__(self, size):
        self.size = size

    @property
    def size(self):
        """(3,) float : The length, width, and height of the box in meters.
        """
        return self._size

    @size.setter
    def size(self, value):
        self._size = np.asanyarray(value).astype(np.float64)


class Cylinder(URDFType):
    """A cylinder whose center is at the local origin.
    Parameters
    ----------
    radius : float
        The radius of the cylinder in meters.
    length : float
        The length of the cylinder in meters.
    """

    _ATTRIBS = {
        'radius': (float, True),
        'length': (float, True),
    }
    _TAG = 'cylinder'

    def __init__(self, radius, length):
        self.radius = radius
        self.length = length

    @property
    def radius(self):
        """float : The radius of the cylinder in meters.
        """
        return self._radius

    @radius.setter
    def radius(self, value):
        self._radius = float(value)

    @property
    def length(self):
        """float : The length of the cylinder in meters.
        """
        return self._length

    @length.setter
    def length(self, value):
        self._length = float(value)


class Sphere(URDFType):
    """A sphere whose center is at the local origin.
    Parameters
    ----------
    radius : float
        The radius of the sphere in meters.
    """
    _ATTRIBS = {
        'radius': (float, True),
    }
    _TAG = 'sphere'

    def __init__(self, radius):
        self.radius = radius

    @property
    def radius(self):
        """float : The radius of the sphere in meters.
        """
        return self._radius

    @radius.setter
    def radius(self, value):
        self._radius = float(value)


class Mesh(URDFType):
    """A triangular mesh object.
    Parameters
    ----------
    filename : str
        The path to the mesh that contains this object. This can be
        relative to the top-level URDF or an absolute path.
    scale : (3,) float, optional
        The scaling value for the mesh along the XYZ axes.
        If ``None``, assumes no scale is applied.
    """
    _ATTRIBS = {
        'filename': (str, True),
        'scale': (np.ndarray, False)
    }
    _TAG = 'mesh'

    def __init__(self, filename, scale=None):
        self.filename = filename
        self.scale = scale

    @property
    def filename(self):
        """str : The path to the mesh file for this object.
        """
        return self._filename

    @filename.setter
    def filename(self, value):

        if value.startswith('package://'):
            value = value.replace('package://', '')

        value = str(abspath / value)

        # print(value)

        self._filename = value

    @property
    def scale(self):
        """(3,) float : A scaling for the mesh along its local XYZ axes.
        """
        return self._scale

    @scale.setter
    def scale(self, value):
        if value is not None:
            value = np.asanyarray(value).astype(np.float64)
        self._scale = value

    @classmethod
    def _from_xml(cls, node, path):
        kwargs = cls._parse(node, path)
        return Mesh(**kwargs)


class Material(URDFType):
    """A material for some geometry.
    Parameters
    ----------
    name : str
        The name of the material.
    color : (4,) float, optional
        The RGBA color of the material in the range [0,1].
    texture : :class:`.Texture`, optional
        A texture for the material.
    """
    _ATTRIBS = {
        'name': (str, True)
    }
    _ELEMENTS = {}
    _TAG = 'material'

    def __init__(self, name, color=None, texture=None):

        # if color is None:
        #     color = name

        self.name = name
        self.color = color
        self.texture = texture

    @classmethod
    def _from_xml(cls, node, path):  # pragma nocover
        kwargs = cls._parse(node, path)

        # Extract the color -- it's weirdly an attribute of a subelement
        color = node.find('color')
        if color is not None:
            color = np.fromstring(
                color.attrib['rgba'], sep=' ', dtype=np.float64)
        kwargs['color'] = color

        return Material(**kwargs)


class Geometry(URDFType):
    """A wrapper for all geometry types.
    Only one of the following values can be set, all others should be set
    to ``None``.
    Parameters
    ----------
    box : :class:`.Box`, optional
        Box geometry.
    cylinder : :class:`.Cylinder`
        Cylindrical geometry.
    sphere : :class:`.Sphere`
        Spherical geometry.
    mesh : :class:`.Mesh`
        Mesh geometry.
    """

    _ELEMENTS = {
        'box': (Box, False, False),
        'cylinder': (Cylinder, False, False),
        'sphere': (Sphere, False, False),
        'mesh': (Mesh, False, False),
    }
    _TAG = 'geometry'

    def __init__(self, box=None, cylinder=None, sphere=None, mesh=None):
        if (box is None and cylinder is None and
                sphere is None and mesh is None):   # pragma nocover
            raise ValueError('At least one geometry element must be set')

        if box is not None:
            self.box = box
            self.ob = rtb.Box(box.size)

        if cylinder is not None:
            self.cylinder = cylinder
            self.ob = rtb.Cylinder(cylinder.radius, cylinder.length)

        if sphere is not None:
            self.sphere = sphere
            self.ob = rtb.Sphere(sphere.radius)

        if mesh is not None:
            self.mesh = mesh
            self.ob = rtb.Mesh(mesh.filename, scale=mesh.scale)

    @property
    def box(self):
        """:class:`.Box` : Box geometry.
        """
        return self._box

    @box.setter
    def box(self, value):
        if value is not None and not isinstance(value, Box):   # pragma nocover
            raise TypeError('Expected Box type')
        self._box = value

    @property
    def cylinder(self):
        """:class:`.Cylinder` : Cylinder geometry.
        """
        return self._cylinder

    @cylinder.setter
    def cylinder(self, value):
        if value is not None and not isinstance(value, Cylinder):
            raise TypeError('Expected Cylinder type')   # pragma nocover
        self._cylinder = value

    @property
    def sphere(self):
        """:class:`.Sphere` : Spherical geometry.
        """
        return self._sphere

    @sphere.setter
    def sphere(self, value):
        if value is not None and not isinstance(value, Sphere):
            raise TypeError('Expected Sphere type')   # pragma nocover
        self._sphere = value

    @property
    def mesh(self):
        """:class:`.Mesh` : Mesh geometry.
        """
        return self._mesh

    @mesh.setter
    def mesh(self, value):
        if value is not None and not isinstance(value, Mesh):
            raise TypeError('Expected Mesh type')   # pragma nocover
        self._mesh = value

    @property
    def geometry(self):   # pragma nocover
        """:class:`.Box`, :class:`.Cylinder`, :class:`.Sphere`, or
        :class:`.Mesh` : The valid geometry element.
        """
        if self.box is not None:
            return self.box
        if self.cylinder is not None:
            return self.cylinder
        if self.sphere is not None:
            return self.sphere
        if self.mesh is not None:
            return self.mesh
        return None


class Collision(URDFType):
    """Collision properties of a link.
    Parameters
    ----------
    geometry : :class:`.Geometry`
        The geometry of the element
    name : str, optional
        The name of the collision geometry.
    origin : (4,4) float, optional
        The pose of the collision element relative to the link frame.
        Defaults to identity.
    """

    _ATTRIBS = {
        'name': (str, False)
    }
    _ELEMENTS = {
        'geometry': (Geometry, True, False),
    }
    _TAG = 'collision'

    def __init__(self, name, origin, geometry):
        self.geometry = geometry
        self.name = name
        self.origin = origin
        self.geometry.ob.base = origin

    @property
    def geometry(self):
        """:class:`.Geometry` : The geometry of this element.
        """
        return self._geometry

    @geometry.setter
    def geometry(self, value):
        if not isinstance(value, Geometry):   # pragma nocover
            raise TypeError('Must set geometry with Geometry object')
        self._geometry = value

    @property
    def name(self):
        """str : The name of this collision element.
        """
        return self._name

    @name.setter
    def name(self, value):
        if value is not None:
            value = str(value)
        self._name = value

    @property
    def origin(self):
        """(4,4) float : The pose of this element relative to the link frame.
        """
        return self._origin

    @origin.setter
    def origin(self, value):
        self._origin = configure_origin(value)

    @classmethod
    def _from_xml(cls, node, path):
        kwargs = cls._parse(node, path)
        kwargs['origin'], _ = parse_origin(node)
        return Collision(**kwargs)


class Visual(URDFType):
    """Visual properties of a link.
    Parameters
    ----------
    geometry : :class:`.Geometry`
        The geometry of the element
    name : str, optional
        The name of the visual geometry.
    origin : (4,4) float, optional
        The pose of the visual element relative to the link frame.
        Defaults to identity.
    material : :class:`.Material`, optional
        The material of the element.
    """
    _ATTRIBS = {
        'name': (str, False)
    }
    _ELEMENTS = {
        'geometry': (Geometry, True, False),
        'material': (Material, False, False)
    }
    _TAG = 'visual'

    def __init__(self, geometry, name=None, origin=None, material=None):
        self.geometry = geometry
        geometry.ob.base = origin
        self.name = name
        self.origin = origin
        self.material = material

        # Do not set material color yet. The top level URDF colors have not
        # been parsed/defined yet so we do not know what 'Grey' or 'Blue2'
        # mean yet. We set colors after these top level definitions come in

        # Do set it if the color was defined in line by the URDF
        if material is not None and material.color is not None:
            self.geometry.ob.color = material.color

    @property
    def geometry(self):
        """:class:`.Geometry` : The geometry of this element.
        """
        return self._geometry

    @geometry.setter
    def geometry(self, value):
        if not isinstance(value, Geometry):   # pragma nocover
            raise TypeError('Must set geometry with Geometry object')
        self._geometry = value

    @property
    def name(self):
        """str : The name of this visual element.
        """
        return self._name

    @name.setter
    def name(self, value):
        if value is not None:
            value = str(value)
        self._name = value

    @property
    def origin(self):
        """(4,4) float : The pose of this element relative to the link frame.
        """
        return self._origin

    @origin.setter
    def origin(self, value):
        self._origin = configure_origin(value)

    @classmethod
    def _from_xml(cls, node, path):
        kwargs = cls._parse(node, path)
        kwargs['origin'], _ = parse_origin(node)
        return Visual(**kwargs)


class Inertial(URDFType):
    """The inertial properties of a link.
    Parameters
    ----------
    mass : float
        The mass of the link in kilograms.
    inertia : (3,3) float
        The 3x3 symmetric rotational inertia matrix.
    origin : (4,4) float, optional
        The pose of the inertials relative to the link frame.
        Defaults to identity if not specified.
    """
    _TAG = 'inertial'

    def __init__(self, mass, inertia, origin=None):
        self.mass = mass
        self.inertia = inertia
        self.origin = origin

    @property
    def mass(self):
        """float : The mass of the link in kilograms.
        """
        return self._mass

    @mass.setter
    def mass(self, value):
        self._mass = float(value)

    @property
    def inertia(self):
        """(3,3) float : The 3x3 symmetric rotational inertia matrix.
        """
        return self._inertia

    @inertia.setter
    def inertia(self, value):
        value = np.asanyarray(value).astype(np.float64)
        if not np.allclose(value, value.T):  # pragma nocover
            raise ValueError('Inertia must be a symmetric matrix')
        self._inertia = value

    @property
    def origin(self):
        """(4,4) float : The pose of the inertials relative to the link frame.
        """
        return self._origin

    @origin.setter
    def origin(self, value):
        self._origin = configure_origin(value)

    @classmethod
    def _from_xml(cls, node, path):
        origin, _ = parse_origin(node)
        mass = float(node.find('mass').attrib['value'])
        n = node.find('inertia')
        xx = float(n.attrib['ixx'])
        xy = float(n.attrib['ixy'])
        xz = float(n.attrib['ixz'])
        yy = float(n.attrib['iyy'])
        yz = float(n.attrib['iyz'])
        zz = float(n.attrib['izz'])
        inertia = np.array([
            [xx, xy, xz],
            [xy, yy, yz],
            [xz, yz, zz]
        ], dtype=np.float64)
        return Inertial(mass=mass, inertia=inertia, origin=origin)


###############################################################################
# Joint types
###############################################################################


class JointCalibration(URDFType):  # pragma nocover
    """The reference positions of the joint.
    Parameters
    ----------
    rising : float, optional
        When the joint moves in a positive direction, this position will
        trigger a rising edge.
    falling :
        When the joint moves in a positive direction, this position will
        trigger a falling edge.
    """
    _ATTRIBS = {
        'rising': (float, False),
        'falling': (float, False)
    }
    _TAG = 'calibration'

    def __init__(self, rising=None, falling=None):
        self.rising = rising
        self.falling = falling

    @property
    def rising(self):
        """float : description.
        """
        return self._rising

    @rising.setter
    def rising(self, value):
        if value is not None:
            value = float(value)
        self._rising = value

    @property
    def falling(self):
        """float : description.
        """
        return self._falling

    @falling.setter
    def falling(self, value):
        if value is not None:
            value = float(value)
        self._falling = value

    def copy(self, prefix='', scale=None):
        """Create a deep copy of the visual with the prefix applied to all names.
        Parameters
        ----------
        prefix : str
            A prefix to apply to all joint and link names.
        Returns
        -------
        :class:`.JointCalibration`
            A deep copy of the visual.
        """
        return JointCalibration(
            rising=self.rising,
            falling=self.falling,
        )


class JointDynamics(URDFType):
    """The dynamic properties of the joint.
    Parameters
    ----------
    damping : float
        The damping value of the joint (Ns/m for prismatic joints,
        Nms/rad for revolute).
    friction : float
        The static friction value of the joint (N for prismatic joints,
        Nm for revolute).
    """
    _ATTRIBS = {
        'damping': (float, False),
        'friction': (float, False),
    }
    _TAG = 'dynamics'

    def __init__(self, damping, friction):
        self.damping = damping
        self.friction = friction

    @property
    def damping(self):    # pragma nocover
        """float : The damping value of the joint.
        """
        return self._damping

    @damping.setter
    def damping(self, value):
        if value is not None:
            value = float(value)
        self._damping = value

    @property
    def friction(self):
        """float : The static friction value of the joint.
        """
        return self._friction

    @friction.setter
    def friction(self, value):
        if value is not None:
            value = float(value)
        self._friction = value


class JointLimit(URDFType):
    """The limits of the joint.
    Parameters
    ----------
    effort : float
        The maximum joint effort (N for prismatic joints, Nm for revolute).
    velocity : float
        The maximum joint velocity (m/s for prismatic joints, rad/s for
        revolute).
    lower : float, optional
        The lower joint limit (m for prismatic joints, rad for revolute).
    upper : float, optional
        The upper joint limit (m for prismatic joints, rad for revolute).
    """

    _ATTRIBS = {
        'effort': (float, True),
        'velocity': (float, True),
        'lower': (float, False),
        'upper': (float, False),
    }
    _TAG = 'limit'

    def __init__(self, effort, velocity, lower=None, upper=None):
        self.effort = effort
        self.velocity = velocity
        self.lower = lower
        self.upper = upper

    @property
    def effort(self):
        """float : The maximum joint effort.
        """
        return self._effort

    @effort.setter
    def effort(self, value):
        self._effort = float(value)

    @property
    def velocity(self):
        """float : The maximum joint velocity.
        """
        return self._velocity

    @velocity.setter
    def velocity(self, value):
        self._velocity = float(value)

    @property
    def lower(self):
        """float : The lower joint limit.
        """
        return self._lower

    @lower.setter
    def lower(self, value):
        if value is not None:
            value = float(value)
        self._lower = value

    @property
    def upper(self):
        """float : The upper joint limit.
        """
        return self._upper

    @upper.setter
    def upper(self, value):
        if value is not None:
            value = float(value)
        self._upper = value


class JointMimic(URDFType):  # pragma nocover
    """A mimicry tag for a joint, which forces its configuration to
    mimic another joint's.
    This joint's configuration value is set equal to
    ``multiplier * other_joint_cfg + offset``.
    Parameters
    ----------
    joint : str
        The name of the joint to mimic.
    multiplier : float
        The joint configuration multiplier. Defaults to 1.0.
    offset : float, optional
        The joint configuration offset. Defaults to 0.0.
    """
    _ATTRIBS = {
        'joint': (str, True),
        'multiplier': (float, False),
        'offset': (float, False),
    }
    _TAG = 'mimic'

    def __init__(self, joint, multiplier=None, offset=None):
        self.joint = joint
        self.multiplier = multiplier
        self.offset = offset

    @property
    def joint(self):
        """float : The name of the joint to mimic.
        """
        return self._joint

    @joint.setter
    def joint(self, value):
        self._joint = str(value)

    @property
    def multiplier(self):
        """float : The multiplier for the joint configuration.
        """
        return self._multiplier

    @multiplier.setter
    def multiplier(self, value):
        if value is not None:
            value = float(value)
        else:
            value = 1.0
        self._multiplier = value

    @property
    def offset(self):
        """float : The offset for the joint configuration
        """
        return self._offset

    @offset.setter
    def offset(self, value):
        if value is not None:
            value = float(value)
        else:
            value = 0.0
        self._offset = value


class SafetyController(URDFType):  # pragma nocover
    """A controller for joint movement safety.
    Parameters
    ----------
    k_velocity : float
        An attribute specifying the relation between the effort and velocity
        limits.
    k_position : float, optional
        An attribute specifying the relation between the position and velocity
        limits. Defaults to 0.0.
    soft_lower_limit : float, optional
        The lower joint boundary where the safety controller kicks in.
        Defaults to 0.0.
    soft_upper_limit : float, optional
        The upper joint boundary where the safety controller kicks in.
        Defaults to 0.0.
    """
    _ATTRIBS = {
        'k_velocity': (float, True),
        'k_position': (float, False),
        'soft_lower_limit': (float, False),
        'soft_upper_limit': (float, False),
    }
    _TAG = 'safety_controller'

    def __init__(self, k_velocity, k_position=None, soft_lower_limit=None,
                 soft_upper_limit=None):
        self.k_velocity = k_velocity
        self.k_position = k_position
        self.soft_lower_limit = soft_lower_limit
        self.soft_upper_limit = soft_upper_limit

    @property
    def soft_lower_limit(self):
        """float : The soft lower limit where the safety controller kicks in.
        """
        return self._soft_lower_limit

    @soft_lower_limit.setter
    def soft_lower_limit(self, value):
        if value is not None:
            value = float(value)
        else:
            value = 0.0
        self._soft_lower_limit = value

    @property
    def soft_upper_limit(self):
        """float : The soft upper limit where the safety controller kicks in.
        """
        return self._soft_upper_limit

    @soft_upper_limit.setter
    def soft_upper_limit(self, value):
        if value is not None:
            value = float(value)
        else:
            value = 0.0
        self._soft_upper_limit = value

    @property
    def k_position(self):
        """float : A relation between the position and velocity limits.
        """
        return self._k_position

    @k_position.setter
    def k_position(self, value):
        if value is not None:
            value = float(value)
        else:
            value = 0.0
        self._k_position = value

    @property
    def k_velocity(self):
        """float : A relation between the effort and velocity limits.
        """
        return self._k_velocity

    @k_velocity.setter
    def k_velocity(self, value):
        self._k_velocity = float(value)


###############################################################################
# Transmission types
###############################################################################


class Actuator(URDFType):
    """An actuator.
    Parameters
    ----------
    name : str
        The name of this actuator.
    mechanicalReduction : str, optional
        A specifier for the mechanical reduction at the joint/actuator
        transmission.
    hardwareInterfaces : list of str, optional
        The supported hardware interfaces to the actuator.
    """
    _ATTRIBS = {
        'name': (str, True),
    }
    _TAG = 'actuator'

    def __init__(self, name, mechanicalReduction=None,
                 hardwareInterfaces=None):
        self.name = name
        self.mechanicalReduction = mechanicalReduction
        self.hardwareInterfaces = hardwareInterfaces

    @property
    def name(self):  # pragma nocover
        """str : The name of this actuator.
        """
        return self._name

    @name.setter
    def name(self, value):
        self._name = str(value)

    @property
    def mechanicalReduction(self):     # pragma nocover
        """str : A specifier for the type of mechanical reduction.
        """
        return self._mechanicalReduction

    @mechanicalReduction.setter
    def mechanicalReduction(self, value):
        if value is not None:
            value = str(value)
        self._mechanicalReduction = value

    @property
    def hardwareInterfaces(self):   # pragma nocover
        """list of str : The supported hardware interfaces.
        """
        return self._hardwareInterfaces

    @hardwareInterfaces.setter
    def hardwareInterfaces(self, value):   # pragma nocover
        if value is None:
            value = []
        else:
            value = list(value)
            for i, v in enumerate(value):
                value[i] = str(v)
        self._hardwareInterfaces = value

    @classmethod
    def _from_xml(cls, node, path):   # pragma nocover
        kwargs = cls._parse(node, path)
        mr = node.find('mechanicalReduction')
        if mr is not None:
            mr = float(mr.text)
        kwargs['mechanicalReduction'] = mr
        hi = node.findall('hardwareInterface')
        if len(hi) > 0:
            hi = [h.text for h in hi]
        kwargs['hardwareInterfaces'] = hi
        return Actuator(**kwargs)


class TransmissionJoint(URDFType):
    """A transmission joint specification.
    Parameters
    ----------
    name : str
        The name of this actuator.
    hardwareInterfaces : list of str, optional
        The supported hardware interfaces to the actuator.
    """
    _ATTRIBS = {
        'name': (str, True),
    }
    _TAG = 'joint'

    def __init__(self, name, hardwareInterfaces):
        self.name = name
        self.hardwareInterfaces = hardwareInterfaces

    @property
    def name(self):     # pragma nocover
        """str : The name of this transmission joint.
        """
        return self._name

    @name.setter
    def name(self, value):
        self._name = str(value)

    @property
    def hardwareInterfaces(self):   # pragma nocover
        """list of str : The supported hardware interfaces.
        """
        return self._hardwareInterfaces

    @hardwareInterfaces.setter
    def hardwareInterfaces(self, value):   # pragma nocover
        if value is None:
            value = []
        else:
            value = list(value)
            for i, v in enumerate(value):
                value[i] = str(v)
        self._hardwareInterfaces = value

    @classmethod
    def _from_xml(cls, node, path):
        kwargs = cls._parse(node, path)
        hi = node.findall('hardwareInterface')
        if len(hi) > 0:
            hi = [h.text for h in hi]
        kwargs['hardwareInterfaces'] = hi
        return TransmissionJoint(**kwargs)


###############################################################################
# Top-level types
###############################################################################


class Transmission(URDFType):
    """An element that describes the relationship between an actuator and a
    joint.
    Parameters
    ----------
    name : str
        The name of this transmission.
    trans_type : str
        The type of this transmission.
    joints : list of :class:`.TransmissionJoint`
        The joints connected to this transmission.
    actuators : list of :class:`.Actuator`
        The actuators connected to this transmission.
    """
    _ATTRIBS = {
        'name': (str, True),
    }
    _ELEMENTS = {
        'joints': (TransmissionJoint, True, True),
        'actuators': (Actuator, True, True),
    }
    _TAG = 'transmission'

    def __init__(self, name, trans_type, joints=None, actuators=None):
        self.name = name
        self.trans_type = trans_type
        self.joints = joints
        self.actuators = actuators

    @property
    def name(self):
        """str : The name of this transmission.
        """
        return self._name

    @name.setter
    def name(self, value):
        self._name = str(value)

    @property
    def trans_type(self):   # pragma nocover
        """str : The type of this transmission.
        """
        return self._trans_type

    @trans_type.setter
    def trans_type(self, value):
        self._trans_type = str(value)

    @property
    def joints(self):     # pragma nocover
        """:class:`.TransmissionJoint` : The joints the transmission is
        connected to.
        """
        return self._joints

    @joints.setter
    def joints(self, value):   # pragma nocover
        if value is None:
            value = []
        else:
            value = list(value)
            for v in value:
                if not isinstance(v, TransmissionJoint):
                    raise TypeError(
                        'Joints expects a list of TransmissionJoint'
                    )
        self._joints = value

    @property
    def actuators(self):     # pragma nocover
        """:class:`.Actuator` : The actuators the transmission is connected to.
        """
        return self._actuators

    @actuators.setter
    def actuators(self, value):   # pragma nocover
        if value is None:
            value = []
        else:
            value = list(value)
            for v in value:
                if not isinstance(v, Actuator):
                    raise TypeError(
                        'Actuators expects a list of Actuator'
                    )
        self._actuators = value

    @classmethod
    def _from_xml(cls, node, path):
        kwargs = cls._parse(node, path)
        if node.find('type') is not None:
            kwargs['trans_type'] = node.find('type').text
        else:
            kwargs['trans_type'] = ' '    # pragma nocover

        return Transmission(**kwargs)


class Joint(URDFType):
    """A connection between two links.
    There are several types of joints, including:
    - ``fixed`` - a joint that cannot move.
    - ``prismatic`` - a joint that slides along the joint axis.
    - ``revolute`` - a hinge joint that rotates about the axis with a limited
      range of motion.
    - ``continuous`` - a hinge joint that rotates about the axis with an
      unlimited range of motion.
    - ``planar`` - a joint that moves in the plane orthogonal to the axis.
    - ``floating`` - a joint that can move in 6DoF.
    Parameters
    ----------
    name : str
        The name of this joint.
    parent : str
        The name of the parent link of this joint.
    child : str
        The name of the child link of this joint.
    joint_type : str
        The type of the joint. Must be one of :obj:`.Joint.TYPES`.
    axis : (3,) float, optional
        The axis of the joint specified in joint frame. Defaults to
        ``[1,0,0]``.
    origin : (4,4) float, optional
        The pose of the child link with respect to the parent link's frame.
        The joint frame is defined to be coincident with the child link's
        frame, so this is also the pose of the joint frame with respect to
        the parent link's frame.
    limit : :class:`.JointLimit`, optional
        Limit for the joint. Only required for revolute and prismatic
        joints.
    dynamics : :class:`.JointDynamics`, optional
        Dynamics for the joint.
    safety_controller : :class`.SafetyController`, optional
        The safety controller for this joint.
    calibration : :class:`.JointCalibration`, optional
        Calibration information for the joint.
    mimic : :class:`JointMimic`, optional
        Joint mimicry information.
    """
    TYPES = ['fixed', 'prismatic', 'revolute',
             'continuous', 'floating', 'planar']
    _ATTRIBS = {
        'name': (str, True),
    }
    _ELEMENTS = {
        'dynamics': (JointDynamics, False, False),
        'limit': (JointLimit, False, False),
        'mimic': (JointMimic, False, False),
        'safety_controller': (SafetyController, False, False),
        'calibration': (JointCalibration, False, False),
    }
    _TAG = 'joint'

    def __init__(self, name, joint_type, parent, child, axis=None, origin=None,
                 limit=None, dynamics=None, safety_controller=None,
                 calibration=None, mimic=None, rpy=None):
        self.name = name
        self.parent = parent
        self.child = child
        self.joint_type = joint_type
        self.axis = axis
        self.origin = origin
        self.rpy = rpy
        self.limit = limit
        self.dynamics = dynamics
        self.safety_controller = safety_controller
        self.calibration = calibration
        self.mimic = mimic

    @property
    def name(self):
        """str : Name for this joint.
        """
        return self._name

    @name.setter
    def name(self, value):
        self._name = str(value)

    @property
    def joint_type(self):
        """str : The type of this joint.
        """
        return self._joint_type

    @joint_type.setter
    def joint_type(self, value):
        value = str(value)
        if value not in Joint.TYPES:   # pragma nocover
            raise ValueError('Unsupported joint type {}'.format(value))
        self._joint_type = value

    @property
    def parent(self):
        """str : The name of the parent link.
        """
        return self._parent

    @parent.setter
    def parent(self, value):
        self._parent = str(value)

    @property
    def child(self):
        """str : The name of the child link.
        """
        return self._child

    @child.setter
    def child(self, value):
        self._child = str(value)

    @property
    def axis(self):
        """(3,) float : The joint axis in the joint frame.
        """
        return self._axis

    @axis.setter
    def axis(self, value):
        if value is None:
            value = np.array([1.0, 0.0, 0.0], dtype=np.float64)
        else:
            value = np.asanyarray(value, dtype=np.float64)
            if value.shape != (3,):    # pragma nocover
                raise ValueError('Invalid shape for axis, should be (3,)')
            norm = np.linalg.norm(value)

            if norm != 0:
                value = value / norm
        self._axis = value

    @property
    def origin(self):
        """(4,4) float : The pose of child and joint frames relative to the
        parent link's frame.
        """
        return self._origin

    @origin.setter
    def origin(self, value):
        self._origin = configure_origin(value)

    @property
    def rpy(self):
        return self._rpy

    @rpy.setter
    def rpy(self, value):
        self._rpy = value

    @property
    def limit(self):
        """:class:`.JointLimit` : The limits for this joint.
        """
        return self._limit

    @limit.setter
    def limit(self, value):
        if value is None:
            if self.joint_type in ['prismatic', 'revolute']:   # pragma nocover
                raise ValueError('Require joint limit for prismatic and '
                                 'revolute joints')
        elif not isinstance(value, JointLimit):   # pragma nocover
            raise TypeError('Expected JointLimit type')
        self._limit = value

    @property
    def dynamics(self):
        """:class:`.JointDynamics` : The dynamics for this joint.
        """
        return self._dynamics

    @dynamics.setter
    def dynamics(self, value):
        if value is not None:
            if not isinstance(value, JointDynamics):   # pragma nocover
                raise TypeError('Expected JointDynamics type')
        self._dynamics = value

    @property
    def safety_controller(self):   # pragma nocover
        """:class:`.SafetyController` : The safety controller for this joint.
        """
        return self._safety_controller

    @safety_controller.setter
    def safety_controller(self, value):
        if value is not None:
            if not isinstance(value, SafetyController):   # pragma nocover
                raise TypeError('Expected SafetyController type')
        self._safety_controller = value

    @property
    def calibration(self):   # pragma nocover
        """:class:`.JointCalibration` : The calibration for this joint.
        """
        return self._calibration

    @calibration.setter
    def calibration(self, value):
        if value is not None:
            if not isinstance(value, JointCalibration):   # pragma nocover
                raise TypeError('Expected JointCalibration type')
        self._calibration = value

    @property
    def mimic(self):   # pragma nocover
        """:class:`.JointMimic` : The mimic for this joint.
        """
        return self._mimic

    @mimic.setter
    def mimic(self, value):
        if value is not None:
            if not isinstance(value, JointMimic):   # pragma nocover
                raise TypeError('Expected JointMimic type')
        self._mimic = value

    def is_valid(self, cfg):    # pragma nocover
        """Check if the provided configuration value is valid for this joint.
        Parameters
        ----------
        cfg : float, (2,) float, (6,) float, or (4,4) float
            The configuration of the joint.
        Returns
        -------
        is_valid : bool
            True if the configuration is valid, and False otherwise.
        """
        if self.joint_type not in ['fixed', 'revolute']:
            return True
        if self.joint_limit is None:
            return True
        cfg = float(cfg)
        lower = -np.infty
        upper = np.infty
        if self.limit.lower is not None:
            lower = self.limit.lower
        if self.limit.upper is not None:
            upper = self.limit.upper
        return (cfg >= lower and cfg <= upper)

    @classmethod
    def _from_xml(cls, node, path):
        kwargs = cls._parse(node, path)
        kwargs['joint_type'] = str(node.attrib['type'])
        kwargs['parent'] = node.find('parent').attrib['link']
        kwargs['child'] = node.find('child').attrib['link']
        axis = node.find('axis')
        if axis is not None:
            axis = np.fromstring(axis.attrib['xyz'], sep=' ')
        kwargs['axis'] = axis
        kwargs['origin'], kwargs['rpy'] = parse_origin(node)
        return Joint(**kwargs)


class Link(URDFType):
    """A link of a rigid object.
    Parameters
    ----------
    name : str
        The name of the link.
    inertial : :class:`.Inertial`, optional
        The inertial properties of the link.
    visuals : list of :class:`.Visual`, optional
        The visual properties of the link.
    collsions : list of :class:`.Collision`, optional
        The collision properties of the link.
    """

    _ATTRIBS = {
        'name': (str, True),
    }
    _ELEMENTS = {
        'inertial': (Inertial, False, False),
        'visuals': (Visual, False, True),
        'collisions': (Collision, False, True),
    }
    _TAG = 'link'

    def __init__(self, name, inertial, visuals, collisions):
        self.name = name
        self.inertial = inertial
        self.visuals = visuals
        self.collisions = collisions

    @property
    def name(self):
        """str : The name of this link.
        """
        return self._name

    @name.setter
    def name(self, value):
        self._name = str(value)

    @property
    def inertial(self):
        """:class:`.Inertial` : Inertial properties of the link.
        """
        return self._inertial

    @inertial.setter
    def inertial(self, value):
        if value is not None and not isinstance(value, Inertial):
            raise TypeError('Expected Inertial object')   # pragma nocover
        # Set default inertial
        if value is None:
            value = Inertial(mass=0.0, inertia=np.eye(3))
        self._inertial = value

    @property
    def visuals(self):
        """list of :class:`.Visual` : The visual properties of this link.
        """
        return self._visuals

    @visuals.setter
    def visuals(self, value):
        if value is None:   # pragma nocover
            value = []
        else:
            value = list(value)
            for v in value:
                if not isinstance(v, Visual):    # pragma nocover
                    raise ValueError('Expected list of Visual objects')
        self._visuals = value

    @property
    def collisions(self):
        """list of :class:`.Collision` : The collision properties of this link.
        """
        return self._collisions

    @collisions.setter
    def collisions(self, value):
        if value is None:    # pragma nocover
            value = []
        else:
            value = list(value)
            for v in value:
                if not isinstance(v, Collision):    # pragma nocover
                    raise ValueError('Expected list of Collision objects')
        self._collisions = value


class URDF(URDFType):
    """The top-level URDF specification.
    The URDF encapsulates an articulated object, such as a robot or a gripper.
    It is made of links and joints that tie them together and define their
    relative motions.
    Parameters
    ----------
    name : str
        The name of the URDF.
    links : list of :class:`.Link`
        The links of the URDF.
    joints : list of :class:`.Joint`, optional
        The joints of the URDF.
    transmissions : list of :class:`.Transmission`, optional
        The transmissions of the URDF.
    materials : list of :class:`.Material`, optional
        The materials for the URDF.
    other_xml : str, optional
        A string containing any extra XML for extensions.
    """
    _ATTRIBS = {
        'name': (str, True),
    }
    _ELEMENTS = {
        'links': (Link, True, True),
        'joints': (Joint, False, True),
        'transmissions': (Transmission, False, True),
        'materials': (Material, False, True),
    }
    _TAG = 'robot'

    def __init__(self, name, links, joints=None,
                 transmissions=None, materials=None,
                 other_xml=None):
        if joints is None:   # pragma nocover
            joints = []
        if transmissions is None:   # pragma nocover
            transmissions = []
        if materials is None:
            materials = []

        # TODO, what does this next line do?
        # why arent the other things validated
        try:
            self._validate_transmissions()
        except Exception:
            pass

        self.name = name
        self.other_xml = other_xml

        # No setters for these
        self._links = list(links)
        self._joints = list(joints)
        self._transmissions = list(transmissions)
        self._materials = list(materials)
        self._material_map = {}

        for x in self._materials:
            if x.name in self._material_map:
                raise ValueError('Two materials with name {} '
                                 'found'.format(x.name))
            self._material_map[x.name] = x

        # check for duplicate names
        if len(self._links) > len(set([x.name for x in self._links])):     # pragma nocover  # noqa
            raise ValueError('Duplicate link names')
        if len(self._joints) > len(set([x.name for x in self._joints])):     # pragma nocover  # noqa
            raise ValueError('Duplicate joint names')
        if len(self._transmissions) > len(
                set([x.name for x in self._transmissions])):     # pragma nocover  # noqa
            raise ValueError('Duplicate transmission names')

        elinks = []
        elinkdict = {}
        # jointdict = {}

        # build the list of links in URDF file order
        for link in self._links:
            elink = rtb.ELink(name=link.name)
            elinks.append(elink)
            elinkdict[link.name] = elink

            # add the inertial parameters
            elink.r = link.inertial.origin[:3, 3]
            elink.m = link.inertial.mass
            elink.inertia = link.inertial.inertia

            # add the visuals to visual list
            try:
                elink.geometry = [v.geometry.ob for v in link.visuals]
            except AttributeError:   # pragma nocover
                pass

            #  add collision objects to collision object list
            try:
                elink.collision = [col.geometry.ob for col in link.collisions]
            except AttributeError:   # pragma nocover
                pass

        # connect the links using joint info
        for joint in self._joints:
            # get references to joint's parent and child
            childlink = elinkdict[joint.child]
            parentlink = elinkdict[joint.parent]

            childlink._parent = parentlink  # connect child link to parent
            childlink._joint_name = joint.name

            # constant part of link transform
            trans = sm.SE3(joint.origin).t
            # TODO, find where reverse is used and change
            # it to [::-1] or do that here
            rot = joint.rpy
            childlink._ets = rtb.ETS.SE3(trans, rot)
            childlink._init_Ts()

            # variable part of link transform
            if joint.joint_type in ('revolute', 'continuous'):   # pragma nocover # noqa
                if joint.axis[0] == 1:
                    var = rtb.ETS.rx()
                elif joint.axis[0] == -1:
                    var = rtb.ETS.rx(flip=True)
                elif joint.axis[1] == 1:
                    var = rtb.ETS.ry()
                elif joint.axis[1] == -1:
                    var = rtb.ETS.ry(flip=True)
                elif joint.axis[2] == 1:
                    var = rtb.ETS.rz()
                elif joint.axis[2] == -1:
                    var = rtb.ETS.rz(flip=True)
            elif joint.joint_type == 'prismatic':   # pragma nocover
                if joint.axis[0] == 1:
                    var = rtb.ETS.tx()
                elif joint.axis[0] == -1:
                    var = rtb.ETS.tx(flip=True)
                elif joint.axis[1] == 1:
                    var = rtb.ETS.ty()
                elif joint.axis[1] == -1:
                    var = rtb.ETS.ty(flip=True)
                elif joint.axis[2] == 1:
                    var = rtb.ETS.tz()
                elif joint.axis[2] == -1:
                    var = rtb.ETS.tz(flip=True)
            elif joint.joint_type == 'fixed':
                var = None

            childlink._v = var

            # joint limit
            try:
                joint.qlim = [joint.limit.lower, joint.limit.upper]
            except AttributeError:
                # no joint limits provided
                pass

            # joint friction
            try:
                if joint.dynamics.friction is not None:
                    childlink.B = joint.dynamics.friction

                # TODO Add damping
                # joint.dynamics.damping
            except AttributeError:
                pass

            # joint gear ratio
            # TODO, not sure if t.joint.name is a thing
            for t in self.transmissions:     # pragma nocover
                if t.name == joint.name:
                    childlink.G = t.actuators[0].mechanicalReduction

            self.elinks = elinks

            # TODO, why did you put the base_link on the end?
            # easy to do it here

    @property
    def name(self):
        """str : The name of the URDF.
        """
        return self._name

    @name.setter
    def name(self, value):
        self._name = str(value)

    @property
    def links(self):
        """list of :class:`.Link` : The links of the URDF.
        This returns a copy of the links array which cannot be edited
        directly. If you want to add or remove links, use
        the appropriate functions.
        """
        return copy.copy(self._links)

    @property
    def link_map(self):   # pragma nocover
        """dict : Map from link names to the links themselves.
        This returns a copy of the link map which cannot be edited
        directly. If you want to add or remove links, use
        the appropriate functions.
        """
        return copy.copy(self._link_map)

    @property
    def joints(self):
        """list of :class:`.Joint` : The links of the URDF.
        This returns a copy of the joints array which cannot be edited
        directly. If you want to add or remove joints, use
        the appropriate functions.
        """
        return copy.copy(self._joints)

    @property
    def joint_map(self):     # pragma nocover
        """dict : Map from joint names to the joints themselves.
        This returns a copy of the joint map which cannot be edited
        directly. If you want to add or remove joints, use
        the appropriate functions.
        """
        return copy.copy(self._joint_map)

    @property
    def transmissions(self):
        """list of :class:`.Transmission` : The transmissions of the URDF.
        This returns a copy of the transmissions array which cannot be edited
        directly. If you want to add or remove transmissions, use
        the appropriate functions.
        """
        return copy.copy(self._transmissions)

    @property
    def transmission_map(self):   # pragma nocover
        """dict : Map from transmission names to the transmissions themselves.
        This returns a copy of the transmission map which cannot be edited
        directly. If you want to add or remove transmissions, use
        the appropriate functions.
        """
        return copy.copy(self._transmission_map)

    @property
    def other_xml(self):   # pragma nocover
        """str : Any extra XML that belongs with the URDF.
        """
        return self._other_xml

    @other_xml.setter
    def other_xml(self, value):
        self._other_xml = value

    @property
    def actuated_joints(self):   # pragma nocover
        """list of :class:`.Joint` : The joints that are independently
        actuated.
        This excludes mimic joints and fixed joints. The joints are listed
        in topological order, starting from the base-most joint.
        """
        return self._actuated_joints

    def _merge_materials(self):
        """Merge the top-level material set with the link materials.
        """
        for link in self.links:
            for v in link.visuals:
                if v.material is None:
                    continue
                if v.material.name in self.material_map:
                    v.material = self._material_map[v.material.name]
                    v.geometry.ob.color = v.material.color
                else:
                    self._materials.append(v.material)
                    self._material_map[v.material.name] = v.material

    @staticmethod
    def load(file_obj):     # pragma nocover
        """Load a URDF from a file.
        Parameters
        ----------
        file_obj : str or file-like object
            The file to load the URDF from. Should be the path to the
            ``.urdf`` XML file. Any paths in the URDF should be specified
            as relative paths to the ``.urdf`` file instead of as ROS
            resources.
        Returns
        -------
        urdf : :class:`.URDF`
            The parsed URDF.
        """
        if isinstance(file_obj, str):
            if os.path.isfile(file_obj):
                parser = ET.XMLParser()
                tree = ET.parse(file_obj, parser=parser)
                path, _ = os.path.split(file_obj)
            else:
                raise ValueError('{} is not a file'.format(file_obj))
        else:
            parser = ET.XMLParser()
            tree = ET.parse(file_obj, parser=parser)
            path, _ = os.path.split(file_obj.name)

        node = tree.getroot()
        return URDF._from_xml(node, path)

    @staticmethod
    def loadstr(str_obj, file_obj):
        """Load a URDF from a file.
        Parameters
        ----------
        file_obj : str or file-like object
            The file to load the URDF from. Should be the path to the
            ``.urdf`` XML file. Any paths in the URDF should be specified
            as relative paths to the ``.urdf`` file instead of as ROS
            resources.
        Returns
        -------
        urdf : :class:`.URDF`
            The parsed URDF.
        """
        if isinstance(str_obj, str):
            if os.path.isfile(file_obj):
                parser = ET.XMLParser()
                bytes_obj = BytesIO(bytes(str_obj, 'utf-8'))
                tree = ET.parse(bytes_obj, parser=parser)
                path, _ = os.path.split(file_obj)

        else:   # pragma nocover
            parser = ET.XMLParser()
            tree = ET.parse(file_obj, parser=parser)
            path, _ = os.path.split(file_obj.name)

        node = tree.getroot()
        return URDF._from_xml(node, path)

    def _validate_transmissions(self):
        """Raise an exception of any transmissions are invalidly specified.
        Checks for the following:
        - Transmission joints have valid joint names.
        """
        for t in self.transmissions:   # pragma nocover
            for joint in t.joints:
                if joint.name not in self._joint_map:
                    raise ValueError('Transmission {} has invalid joint name '
                                     '{}'.format(t.name, joint.name))

    @classmethod
    def _from_xml(cls, node, path):
        valid_tags = set([
            'joint', 'link', 'transmission',
            'material'
            ])
        kwargs = cls._parse(node, path)

        extra_xml_node = ET.Element('extra')
        for child in node:
            if child.tag not in valid_tags:
                extra_xml_node.append(child)

        # data = ET.ostring(extra_xml_node)
        # kwargs['other_xml'] = data
        return URDF(**kwargs)
